
//		PrinterStuff.cpp - Printers plugin for Take Command


#define UNICODE					1
#define _UNICODE				1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>
#include <winspool.h>

#include "Plugin.h"
#include "TakeCmd.h"
#include "tcpxcv.h"


#define VER_MAJOR				0
#define VER_MINOR				70
#define VER_BUILD				0
#define VER_PATCH				4

#define PLUGIN_NAME				L"PrinterStuff"

#define ERRHANDLE				GetStdHandle( STD_ERROR_HANDLE )


#ifdef	_WIN64

#define PLATFORM				L"x64"

#else	//	_WIN64

#define PLATFORM				L"x86"

#endif	//	_WIN64



const wchar_t AddIPPortHelpText[] = L"Create a local IP printer port.\n\
\n\
ADDIPPORT address /P /Q\r\n\
\n\
\t/P:port    port name\n\
\t/Q:queue   LPR queue name\n\
\n\
This address is required.  It will usually be a static IP, but may be a\n\
hostname instead.\n";


const wchar_t AddPrinterHelpText[] = L"Install a new local printer.\n\
\n\
ADDPRINTER printername drivername /A /C /D /I /L /O /P /Q /R /T\r\n\
\n\
\t/A:address    host address\n\
\t/C:string     printer comment\n\
\t/D            make the new printer the default\n\
\t/I:filename   INF filename\n\
\t/L:string     printer location\n\
\t/O            only install if it doesn\u2019t already exist\n\
\t/P:port       port name\n\
\t/Q:queue      LPR queue name\n\
\t/R:filename   read printer settings from a file\n\
\t/T            send a test page\n\
\n\
The printername and drivername are both required.  Specify either /A or\n\
/P but not both (/A automatically creates an IP port.)  The filename may\n\
contain <Win> or <x64> tags to select the appropriate driver for the\n\
target system.  Quote any parameter containing spaces or other special\n\
characters.\n";


const wchar_t DefaultPrintHelpText[] = L"Display or change the default printer.\n\
\n\
DEFAULTPRINT name\r\n\
\n\
If name is not specified, the current default printer will be shown.\n";


const wchar_t DelIPPortHelpText[] = L"Delete local IP printing ports.\n\
\n\
DELIPPORT /N /Y name...\r\n\
\n\
\t/N  not really\n\
\t/Y  yes to all prompts\n\
\n\
You must provide at least one name.  This command only deletes IP\n\
printing ports, and only on the local machine.  It will not delete\n\
other kinds of printer ports.  Quote any names which contain spaces.\n";


const wchar_t DelPrinterHelpText[] = L"Delete printers from the local machine.\n\
\n\
DELPRINTER /K /N /Y name...\r\n\
\n\
\t/K  also remove local IP ports\n\
\t/N  not really\n\
\t/Y  yes to all prompts\n\
\n\
You must provide at least one name.  This command deletes printers only\n\
from the system where it is run.  If a network printer is specifed, the\n\
connection will be removed.  Quote any name which contains spaces.\n";


const wchar_t KillJobsHelpText[] = L"Cancel print jobs.\n\
\n\
KILLJOBS /P:printer job...\r\n\
\n\
If no printer is specified, the default printer will be assumed.  The\n\
jobs may be job names, possibly including wildcards; or job ID numbers,\n\
as listed by the LISTJOBS command.  If no jobs are specified, the first\n\
job in the printer\u2019s queue will be canceled.  To cancel all jobs in\n\
the queue, use an asterisk.\n";


const wchar_t ListJobsHelpText[] = L"List print jobs.\n\
\n\
LISTJOBS /P:printer job...\r\n\
\n\
If no printer is specified, jobs for the default printer will be listed.\n\
The \u2018job\u2019 parameters may be job names, possibly including wildcards; or\n\
job ID numbers.\n";


const wchar_t ListPortsHelpText[] = L"List printer ports.\n\
\n\
LISTPORTS /I /N name...\r\n\
\n\
\t/I:  list IP ports\n\
\t/N:  list non-IP ports\n\
\n\
If neither /I nor /N is specified, both IP and non-IP printer ports\n\
will be listed.  Names may contain wildcards.\n";


const wchar_t ListPrintersHelpText[] = L"List installed printers.\n\
\n\
LISTPRINTERS /I /L /N /V name...\r\n\
\n\
\t/I:  local IP printers\n\
\t/L:  local printers\n\
\t/N:  network printers\n\
\t/V   verbose\n\
\n\
If none of /I /L or /N is specified, both local and network printers\n\
will be listed.  Names may contain wildcards.\n";


const wchar_t PrinterDataHelpText[] = L"Save or restore printer settings.\n\
\n\
PRINTERDATA printername /S:filename /R:filename\r\n\
\n\
\t/S:  save printer settings\n\
\t/R:  restore printer settings\n\
\n\
You must specify either /S: or /R:.  Quote any parameter that contains\n\
spaces.\n\n\
Settings saved with /S: can only be restored to the same printer.\n";


const wchar_t PrintPrefsHelpText[] = L"Display or change printing defaults.\n\
\n\
PRINTPREFS /C:n /D:n /O:n /P:n printername\r\n\
\n\
\t/C:  color mode\n\
\t/D:  duplex\n\
\t/O:  orientation\n\
\t/P:  paper size\n\
\n\
If you do not specify a printername, the default printer will\n\
be assumed.\n";


const wchar_t PtrStuffHelpHelpText[] = L"Open the PrinterStuff plugin help file.\n\
\n\
PTRSTUFFHELP topic\r\n\
\n\
Opens this plugin\u2019s help file to the specified topic.\n\
If no topic is given, opens the Overview page by default.  \n";


const wchar_t SetPrinterHelpText[] = L"Set a printer\u2019s status.\n\
\n\
SETPRINTER printername /OF /ON /P /PU /R\r\n\
\n\
\t/OF   take the printer offline\n\
\t/ON   bring the printer back online\n\
\t/P    pause the printer\n\
\t/PU   purge the print queue\n\
\t/R    restart a paused printer\n\
\n\
Quote the printer name if it contains spaces.  If printername is not\n\
specified, the default printer is assumed.\n";


const wchar_t TestPageHelpText[] = L"Print a test page.\n\
\n\
TESTPAGE printername\r\n\
\n\
Quote the printer name if it contains spaces.  If printername is not\n\
specified, the default printer is assumed.\n";



const int NumHelpTexts			= 14;

const wchar_t HelpFeature[NumHelpTexts][13]		= {
	L"addipport", L"addprinter", L"defaultprint", L"delipport", L"delprinter",
	L"killjobs", L"listjobs", L"listports", L"listprinters", L"printerdata",
	L"printprefs", L"ptrstuffhelp", L"setprinter", L"testpage" };

const wchar_t *HelpText[NumHelpTexts]			= {
	AddIPPortHelpText, AddPrinterHelpText, DefaultPrintHelpText, DelIPPortHelpText, DelPrinterHelpText,
	KillJobsHelpText, ListJobsHelpText, ListPortsHelpText, ListPrintersHelpText, PrinterDataHelpText,
	PrintPrefsHelpText, PtrStuffHelpHelpText, SetPrinterHelpText, TestPageHelpText };



const wchar_t PluginFeatures[] = L"_defprinter,_numports,_numprinters,addipport,addprinter,defaultprint,delipport,delprinter,\
killjobs,listjobs,listports,listprinters,printerdata,printprefs,ptrstuffhelp,setprinter,testpage,@addipport,@numports,\
@numprinters,@ptrcomment,@ptrdriver,@ptrexist,@ptrhost,@ptrip,@ptrjobs,@ptrloc,@ptrname,@ptroffline,@ptrport,@ptrserver,\
@ptrshare";



const wchar_t NadaString[]  = L"N/A";
const wchar_t ErrorString[] = L"???";
const wchar_t DefaultQueueName[] = L"raw";


#define SHORT_BUF_LEN		1024
#define MAX_ARG_LEN			8192
#define MAX_ARGS			1024
#define MAX_FILENAME_LEN	16384

LPTSTR arg[MAX_ARG_LEN];
int arglen[MAX_ARG_LEN];
int numargs = 0;
wchar_t allargs[MAX_ARG_LEN] = L"";



HINSTANCE MyDLLHandle = NULL;



BYTE * PrintersList = NULL;
DWORD NumPrinters = 0;

PORT_INFO_2 * PortsList = NULL;
DWORD NumPorts = 0;

BYTE * PrinterInfo = NULL;

JOB_INFO_1 * PrintJobs = NULL;




int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD dwReason, LPVOID lpReserved )
{
	MyDLLHandle = (HINSTANCE) hModule;

	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



#include "NewHelp.cpp"



#define PARSE_BREAK_SPACES		1
#define PARSE_BREAK_COMMAS		2
#define PARSE_SLASHES_KLUDGE	4
#define	PARSE_QUOTES_KLUDGE		8
#define PARSE_FORCE_UPPERCASE	128


int ParseArgs( LPTSTR inString, int flags )
{
	/*
		flags:  bit 0 - break args at spaces
				bit 1 - break args at commas (either bit 0 or bit 1 must be set)
				bit 2 - slashes kludge:   treat /a/b like /a /b
				bit 3 - quotes kludge:    treat /a"foo" like /a:"foo"
				bit 7 - force uppercase
	*/

	if ( !inString )
		return -1;

	numargs = 0;

	int argnum = 0;
	int i = 0;
	int a = 0;
	int b = 0;
	wchar_t ch = '\0';
	BOOL quote  = FALSE;
	BOOL adv    = FALSE;
	BOOL done   = FALSE;

	if ( ( flags & 3 ) == 0 )
		flags |= 1;

	for ( i = 0; i < MAX_ARG_LEN; i++ ) {
		arg[i] = NULL;
		arglen[i] = 0;
	}

	for ( i = 0; !done; i++ ) {
		ch = inString[i];
		adv = FALSE;

		if ( ch == '"' ) {
			quote = !quote;

			if ( quote && ( a == 2 ) && ( arg[argnum][0] == '/' ) && ( iswalpha( arg[argnum][1] ) ) && ( flags & 8 ) )
				ch = ':';
			else
				continue;
		}

		if ( iswspace( ch ) && ( a == 0 ) )
			continue;

		if ( iswspace( ch ) && !quote && ( flags & 1 ) )
			adv = TRUE;

		if ( ( ch == ',' ) && !quote && ( flags & 2 ) )
			adv = TRUE;

		if ( ( ch == '/' ) && !quote && ( a > 1 ) && ( arg[argnum][0] == '/' ) && ( flags & 4 ) ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
		}

		if ( ch == '\0' ) {
			if ( a )
				argnum++;

			done = TRUE;
			continue;
		}

		if ( adv ) {
			argnum++;
			a = 0;
			b++;
			allargs[b] = '\0';
			continue;
		}

		if ( b >= ( MAX_ARG_LEN - 1 ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Buffer overflow: b = %d, MAX_ARG_LEN = %d\r\n", b, MAX_ARG_LEN );
			return -1;
		}

		if ( argnum < MAX_ARG_LEN ) {
			if ( a == 0 )
				arg[argnum] = allargs + b;

			if ( flags & 128 )
				ch = towupper( ch );

			a++;
			arglen[argnum]  = a;
			allargs[b++] = ch;
			allargs[b]   = '\0';
		}

	}

	for ( i = 0; i < argnum; i++ ) {
		while ( ( arglen[i] > 0 ) && iswspace( arg[i][arglen[i] - 1] ) ) {
			arglen[i]--;
			arg[i][arglen[i]] = '\0';
		}
	}

	numargs = argnum;
	return argnum;
}



long int GetInt( LPTSTR inString )
{
	if ( !inString )
		return -1;

	long int rv = 0;
	wchar_t *sep1 = '\0';

	if ( !iswdigit( inString[0] ) )
		return -1;

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) )
		rv = wcstol( inString + 2, &sep1, 16 );
	else
		rv = wcstol( inString, &sep1, 10 );

	if ( *sep1 != '\0' )
		return -1;

	return rv;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}


 
int FreePrintersList()
{
	if ( PrintersList ) {
		free( PrintersList );
		PrintersList = NULL;
	}

	NumPrinters = 0;
	return 0;
}



int GetPrintersList( int EnumLevel, DWORD EnumFlags )
{
	DWORD BufferSize = 0;
	DWORD BytesNeeded = 0;

	NumPrinters = 0;

	FreePrintersList();


	//  Find the size required for the buffer:

	EnumPrinters( EnumFlags, NULL, EnumLevel, NULL, 0, &BytesNeeded, &NumPrinters );

	if ( !BytesNeeded )
		return -1;

	// Printf( L"Bytes needed: %d\r\n", BytesNeeded );


	//  Allocate memory for the buffer:

	PrintersList = (BYTE*) malloc( BytesNeeded );

	if ( !PrintersList )
		return ERROR_NOT_ENOUGH_MEMORY;


	//  Get list of printers:

	if ( !EnumPrinters( EnumFlags, NULL, EnumLevel, PrintersList, BytesNeeded, &BytesNeeded, &NumPrinters ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error enumerating printers\r\n" );
		return -1;
	}
	// Printf( L"EnumPrinters() lists %d printers.\r\n", NumPrinters );

	return 0;
}



int FreePortsList()
{
	if ( PortsList ) {
		free( PortsList );
		PortsList = NULL;
	}

	NumPorts = 0;
	return 0;
}



int GetPortsList( void )
{
	DWORD BufferSize		= 0;
	DWORD BytesNeeded		= 0;
	NumPorts				= 0;

	FreePortsList();


	//  Find the size required for the buffer:

	EnumPorts( NULL, 2, NULL, 0, &BytesNeeded, &NumPorts );

	if ( !BytesNeeded )
		return -1;


	//  Allocate memory for the buffer:

	PortsList = (PORT_INFO_2*) malloc( BytesNeeded );

	if ( !PortsList )
		return ERROR_NOT_ENOUGH_MEMORY;

	
	//  Get list of printer ports:

	EnumPorts( NULL, 2, (LPBYTE) PortsList, BytesNeeded, &BytesNeeded, &NumPorts );

	return 0;
}



BOOL IsIPPort( LPTSTR PortName )
{
	if ( !PortName || !PortsList || !NumPorts )
		return FALSE;

	int PortIndex = -1;

	for ( unsigned int i = 0; i < NumPorts; i++ )
		if ( !_wcsicmp( PortsList[i].pPortName, PortName ) )
			PortIndex = i;

	if ( PortIndex < 0 )
		return FALSE;

	return ( ( PortsList[PortIndex].fPortType & PORT_TYPE_NET_ATTACHED ) != 0 );
}



BOOL PortNameInList( LPTSTR PortName )
{
	if ( !PortName || !PortsList || !NumPorts )
		return FALSE;


	for ( unsigned int i = 0; i < NumPorts; i++ )
		if ( !_wcsicmp( PortName, PortsList[i].pPortName ) )
			return TRUE;

	return FALSE;
}



int FreePrinterInfo()
{
	if ( PrinterInfo ) {
		free( PrinterInfo );
		PrinterInfo = NULL;
	}

	return 0;
}



int GetPrinterInfo( HANDLE PrinterHandle, int InfoLevel )
{
	if ( !PrinterHandle )
		return -1;

	DWORD BytesNeeded = 0;
	int rv = -1;

	FreePrinterInfo();


	//  Get the size of the buffer required for the requested printer info:

	GetPrinter( PrinterHandle, InfoLevel, NULL, 0, &BytesNeeded );


	//  Allocate memory for the buffer:

	PrinterInfo = (BYTE*) malloc( BytesNeeded );


	//  Get the printer info:

	if ( GetPrinter( PrinterHandle, InfoLevel, PrinterInfo, BytesNeeded, &BytesNeeded ) ) {
		rv = 0;
		// Printf( L"Second GetPrinter() returned happy.\r\n" );
	}

	return rv;
}



int GetPrinterInfo( LPTSTR PrinterName, int InfoLevel )
{
	if ( !PrinterName )
		return -1;

	HANDLE PrinterHandle = NULL;
	//  PRINTER_DEFAULTS Defaults = { 0, 0, PRINTER_ACCESS_USE };
	DWORD BytesNeeded = 0;
	int rv = -1;

	FreePrinterInfo();


	//  Open a handle to the printer:

	if ( !OpenPrinter( PrinterName, &PrinterHandle, NULL ) )
		return -1;
	// Printf( L"PrinterHandle is %x.\r\n", PrinterHandle );


	//  Get the size of the buffer required for the requested printer info:

	GetPrinter( PrinterHandle, InfoLevel, NULL, 0, &BytesNeeded );


	//  Allocate memory for the buffer:

	PrinterInfo = (BYTE*) malloc( BytesNeeded );


	//  Get the printer info:

	if ( GetPrinter( PrinterHandle, InfoLevel, PrinterInfo, BytesNeeded, &BytesNeeded ) ) {
		rv = 0;
		// Printf( L"Second GetPrinter() returned happy.\r\n" );
	}

	ClosePrinter( PrinterHandle );

	return rv;
}



DWORD GetPrinterAttribs( HANDLE PrinterHandle )
{
	FreePrinterInfo();

	DWORD rv = 0;
	DWORD BytesNeeded = 0;
	PRINTER_INFO_2 * Info2;
	int Attribs = 0;

	GetPrinter( PrinterHandle, 2, NULL, 0, &BytesNeeded );
	PrinterInfo = (BYTE*) malloc( BytesNeeded );

	if ( GetPrinter( PrinterHandle, 2, PrinterInfo, BytesNeeded, &BytesNeeded ) ) {
		Info2 = (PRINTER_INFO_2*) PrinterInfo;
		rv = Info2->Attributes;
	}

	FreePrinterInfo();
	return rv;
}



int RemoveIPPort( LPTSTR PortName )
{
	if ( !PortName )
		return -1;
	
	if ( PortName[0] == '\0' )
		return -1;

	if ( wcslen( PortName ) > MAX_PORTNAME_LEN )
		return -1;


	DELETE_PORT_DATA_1 Target;

	memset( &Target, 0, sizeof( Target ) );
	wcscpy_s( Target.psztPortName, MAX_PORTNAME_LEN, PortName );
	Target.dwVersion = 1;

	HANDLE hXcv = INVALID_HANDLE_VALUE;
	PRINTER_DEFAULTS Defaults = { NULL, NULL, SERVER_ACCESS_ADMINISTER };

	DWORD BytesNeeded = 0;
	DWORD XcvStatus = 0;
	BOOL rv = FALSE;
	int err = 0;


	if ( !OpenPrinter( L",XcvMonitor Standard TCP/IP Port", &hXcv, &Defaults ) ) {
		err = GetLastError();
		if ( !err )
			err = -1;

		return err;
	}

	rv = XcvData( hXcv, L"DeletePort", (PBYTE) &Target, sizeof( Target ), NULL, 0, &BytesNeeded, &XcvStatus );

	ClosePrinter( hXcv );

	if ( !rv || ( XcvStatus != NO_ERROR ) ) {
		err = GetLastError();
		if ( !err )
			err = -1;

		return err;
	}

	return 0;
}



BOOL LooksLikeAnIP( LPTSTR inString )
{
	if ( !inString )
		return FALSE;

	int octet = 0, i = 0;
	int val = 0;

	do {
		if ( !iswdigit( inString[i] ) )
			return FALSE;
		else
			val = inString[i++] - '0';

		if ( iswdigit( inString[i] ) )
			val = ( val * 10 ) + ( inString[i++] - '0' );

		if ( iswdigit( inString[i] ) )
			val = ( val * 10 ) + ( inString[i++] - '0' );

		if ( inString[i++] != ( ( octet == 3 ) ? '\0' : '.' ) )
			return FALSE;

		if ( val > 255 )
			return FALSE;

		octet++;

	} while ( octet < 4 );

	return TRUE;
}



int GeneratePortName( LPTSTR HostAddress, LPTSTR PortName )
{
	if ( !HostAddress || !PortName )
		return -1;

	wchar_t shortbuf[SHORT_BUF_LEN] = L"IP__";

	if ( !LooksLikeAnIP( HostAddress ) ) {

		size_t i = 0, j = wcslen( shortbuf );
		wchar_t ch = '\0';

		do {
			ch = HostAddress[i++];

			if ( j < ( SHORT_BUF_LEN - 1 ) ) {
				shortbuf[j++] = towupper( ch );
				shortbuf[j] = '\0';
			}

		} while ( ch != '\0' );

		if ( j < ( SHORT_BUF_LEN - 1 ) ) {
			shortbuf[j - 1] = '$';
			shortbuf[j] = '\0';
		}

		wcscpy_s( PortName, SHORT_BUF_LEN, shortbuf );
		return 0;
	}
	

	int val = 0, hi = 0, lo = 0;
	size_t i = 0, j = wcslen( shortbuf );
	wchar_t ch = '\0', hdig = '\0', ldig = '\0' ;

	do {
		ch = HostAddress[i++];

		if ( iswdigit( ch ) )
			val = ( 10 * val ) + ( ch - '0' );
		else {
			hi = val / 16;
			lo = val & 15;

			if ( hi < 10 )
				hdig = '0' + hi;
			else
				hdig = 'A' + hi - 10;

			if ( lo < 10 )
				ldig = '0' + lo;
			else
				ldig = 'A' + lo - 10;

			shortbuf[j++] = hdig;
			shortbuf[j++] = ldig;

			val = 0;
		}

	} while ( ch );

	shortbuf[j++] = '$';
	shortbuf[j] = '\0';

	wcscpy_s( PortName, SHORT_BUF_LEN, shortbuf );
	return 0;
}



int CreateIPPort( LPTSTR PortName, LPTSTR HostName, LPTSTR IPAddress, LPTSTR QueueName, int Protocol, int PortNumber )
{
	//  Check inputs:

	if ( !PortName )
		return -1;

	if ( wcslen( PortName ) >= MAX_PORTNAME_LEN )
		return -1;

	if ( !HostName && !IPAddress )
		return -1;



	wchar_t temp[SHORT_BUF_LEN]		= L"";

	size_t len					= 0;
	int err						= 0;
	int j						= 0;
	wchar_t *sep1				= NULL;
	int rv						= 0;

	DWORD BytesNeeded			= 0;
	PORT_DATA_1 PortData;
	HANDLE hXcv					= INVALID_HANDLE_VALUE;
	PRINTER_DEFAULTS Defaults	= { NULL, NULL, SERVER_ACCESS_ADMINISTER };
	DWORD XcvStatus				= 0;


	//		Get list of existing ports :

	if ( rv = GetPortsList() )
		return rv;


	//		Generate default port name, if none was supplied :

	if ( HostName )
		if ( HostName[0] )
			GeneratePortName( HostName, temp );

	if ( temp[0] == '\0' )
		if ( IPAddress )
			if ( IPAddress[0] )
				GeneratePortName( IPAddress, temp );

	if ( PortName[0] == '\0' )
		wcscpy_s( PortName, SHORT_BUF_LEN, temp );

	sep1 = wcsrchr( temp, '$' );
	if ( sep1 )
		*sep1 = '\0';


	//		If port name matches an existing port, generate a unique name for the new port :

	while ( PortNameInList( PortName ) )
		wsprintf( PortName, L"%s_%u$", temp, ++j );


	//		Check host and IP address name lengths :

	if ( HostName ) {
		len += wcslen( HostName );

		if ( wcslen( HostName ) >= MAX_NETWORKNAME_LEN )
			goto ErrorExit;
	}

	if ( IPAddress ) {
		len += wcslen( IPAddress );

		if ( wcslen( IPAddress ) >= MAX_IPADDR_STR_LEN )
			goto ErrorExit;
	}

	if ( !len )
		goto ErrorExit;


	//		Sanity check protocol and TCP/IP port :

	if ( !Protocol )
		Protocol = PROTOCOL_RAWTCP_TYPE;

	if ( ( Protocol != PROTOCOL_RAWTCP_TYPE ) && ( Protocol != PROTOCOL_LPR_TYPE ) )
		goto ErrorExit;

	if ( !PortNumber ) {
		if ( Protocol == PROTOCOL_RAWTCP_TYPE )
			PortNumber = 9100;
		else
			PortNumber = 515;
	}


	//		Set up PORT_INFO_1 structure for creation of a new IP printing port :

	memset( &PortData, 0, sizeof( PORT_DATA_1 ) );

	wcscpy_s( PortData.sztPortName, MAX_PORTNAME_LEN, PortName );
	PortData.dwVersion		= 1;
	PortData.dwProtocol		= Protocol;
	PortData.cbSize			= sizeof( PortData );
	PortData.dwPortNumber	= PortNumber;
	PortData.dwSNMPDevIndex	= 1;
	wcscpy_s( PortData.sztSNMPCommunity, MAX_SNMP_COMMUNITY_STR_LEN, L"public" );

	if ( !HostName )
		wcscpy_s( PortData.sztHostAddress, MAX_NETWORKNAME_LEN, IPAddress );
	else if ( HostName[0] == '\0' )
		wcscpy_s( PortData.sztHostAddress, MAX_NETWORKNAME_LEN, IPAddress );
	else
		wcscpy_s( PortData.sztHostAddress, MAX_NETWORKNAME_LEN, HostName );

	if ( IPAddress )
		wcscpy_s( PortData.sztIPAddress, MAX_IPADDR_STR_LEN, IPAddress );

	if ( Protocol == PROTOCOL_LPR_TYPE ) {
		if ( !QueueName )
			wsprintf( PortData.sztQueue, DefaultQueueName );
		else if ( QueueName[0] == '\0' )
			wsprintf( PortData.sztQueue, DefaultQueueName );
		else
			wcscpy_s( PortData.sztQueue, MAX_QUEUENAME_LEN, QueueName );
	}


	//  Create new IP port:

	if ( OpenPrinter( L",XcvMonitor Standard TCP/IP Port", &hXcv, &Defaults ) ) {

		if ( !XcvData( hXcv, L"AddPort", (PBYTE) &PortData, sizeof( PortData ), NULL, 0, &BytesNeeded, &XcvStatus ) ) {
			err = GetLastError();
			goto ErrorExit;
		}

		ClosePrinter( hXcv );
	}
	else {
		err = GetLastError();
		goto ErrorExit;
	}


	//		Success!

	FreePortsList();
	return 0;


ErrorExit:

	//		Failure!

	if ( !err )
		err = -1;

	if ( hXcv && ( hXcv != INVALID_HANDLE_VALUE ) )
		ClosePrinter( hXcv );

	FreePortsList();
	return err;
}



DLLExports int WINAPI _defprinter( LPTSTR inString )
{
	if ( !inString )
		return 1;

	DWORD MaxSize = SHORT_BUF_LEN;

	inString[0] = '\0';

	GetDefaultPrinter( inString, &MaxSize );

	if ( inString[0] == '\0' )
		wcscpy_s( inString, SHORT_BUF_LEN, NadaString );

	return 0;
}



DLLExports int WINAPI _numports( LPTSTR inString )
{
	if ( !inString )
		return 1;

	FreePortsList();


	//  Get list of printer ports:

	if ( GetPortsList() )
		wcscpy_s( inString, SHORT_BUF_LEN, L"0" );
	else
		//  Write number of ports to output string:
		Sprintf( inString, L"%d", NumPorts );


	//  Release the buffer:

	FreePortsList();
	return 0;
}



int GetErrorText( int err, LPTSTR outString )
{
	if ( !outString )
		return -1;

	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( !FormatMessage( FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK, NULL, err, 0, outString, SHORT_BUF_LEN, NULL ) )
		Sprintf( outString, L"Error %d", err );

	size_t l = wcslen( outString );

	if ( l > 0 )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



DLLExports int WINAPI delipport( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	BYTE argtype[MAX_ARG_LEN];

	BOOL ExcludeName    = FALSE;
	BOOL IncludeName    = FALSE;
	BOOL ExactMatch     = FALSE;
	BOOL AnyPortSpecs   = FALSE;
	BOOL AnySpecsAtAll  = FALSE;
	BOOL NotReally      = FALSE;
	BOOL NoPrompts      = FALSE;
	BOOL DoIt           = FALSE;

	int PortsFound = 0;
	int PortsDeleted = 0;
	int rv = 0;
	int errorlevel = 0;

	wchar_t ErrorMessage[SHORT_BUF_LEN] = L"";



	for ( int i = 0; i < numargs; i++ ) {
		argtype[i] = 0;

		if ( arglen[i] ) {

			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( DelIPPortHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( DelIPPortHelpText, TRUE );
			else if ( !_wcsicmp( arg[i], L"/N" ) )
				NotReally = TRUE;
			else if ( !_wcsicmp( arg[i], L"/Y" ) )
				NoPrompts = TRUE;
			else if ( ( arg[i][0] == '!' ) && ( arg[i][1] != '\0' ) ) {
				argtype[i] = 2;
				AnySpecsAtAll = TRUE;
			}
			else if ( arg[i][0] != '/' ) {
				argtype[i] = 1;
				AnySpecsAtAll = AnyPortSpecs = TRUE;
			}
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}

	if ( !AnySpecsAtAll ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing port name\r\n" );
		return 1;
	}

	FreePortsList();


	//  Get list of ports:

	if ( GetPortsList() ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t get list of printer ports\r\n" );
		return 2;
	}


	//  Loop through list of ports:

	for ( unsigned int i = 0; i < NumPorts; i++ ) {

		if ( !IsIPPort( PortsList[i].pPortName ) )
			continue;

		ExcludeName = FALSE;
		IncludeName = !AnyPortSpecs;
		ExactMatch  = FALSE;

		for ( int j = 0; j < numargs; j++ ) {
			if ( argtype[j] == 2 ) {
				if ( !WildcardComparison( arg[j]+1, PortsList[i].pPortName, 0, 1 ) )
					ExcludeName = TRUE;
			}
			else if ( argtype[j] == 1 ) {
				if ( !_wcsicmp( arg[j], PortsList[i].pPortName ) )
					ExactMatch = TRUE;
				else if ( !WildcardComparison( arg[j], PortsList[i].pPortName, 0, 1 ) )
					IncludeName = TRUE;
			}
		}

		if ( !ExactMatch )
			if ( ExcludeName || !IncludeName )
				continue;

		DoIt = TRUE;

		if ( !NoPrompts && !ExactMatch ) {
			//  Prompt the user:

			Printf( L"Delete IP printing port \"%s\"", PortsList[i].pPortName );
			int k = QueryInputChar( L"", L"Y/N/A" );
			
			switch ( k ) {
				case 'A':
					NoPrompts = TRUE;
					// fall through
				case 'Y':
					DoIt = TRUE;
					break;
				default:
					DoIt = FALSE;
			}
		}

		PortsFound++;

		if ( DoIt ) {
			if ( NotReally ) {
				Printf( L"IP printing port \"%s\" deleted.\r\n", PortsList[i].pPortName );
				PortsDeleted++;
			}
			else {
				rv = RemoveIPPort( PortsList[i].pPortName );

				if ( !rv ) {
					Printf( L"IP printing port \"%s\" deleted.\r\n", PortsList[i].pPortName );
					PortsDeleted++;
				}
				else {
					DisplayErrorHeader();
					GetErrorText( rv, ErrorMessage );
					Qprintf( ERRHANDLE, L"Can\u2019t delete IP printing port \"%s\" (%s)\r\n", PortsList[i].pPortName, ErrorMessage );
					errorlevel = 2;
				}
			}
		}

		QPuts( L"\r\n" );
	}


	//  Show summary report:

	if ( !PortsFound )
		QPuts( L"No matching IP printing ports found.\r\n" );
	else
		Printf( L"%d matching IP printing %s found, %d %s.\r\n", PortsFound, ( PortsFound == 1 ) ? L"port" : L"ports", PortsDeleted, NotReally ? L"would be deleted" : L"deleted" );


	//  Clean up and exit:

	FreePortsList();
	return errorlevel;
}



DLLExports int WINAPI _numprinters( LPTSTR inString )
{
	if ( !inString )
		return 1;

	FreePrintersList();


	//  Get list of printers:

	if ( GetPrintersList( 4, PRINTER_ENUM_LOCAL | PRINTER_ENUM_CONNECTIONS ) )
		wcscpy_s( inString, SHORT_BUF_LEN, L"0" );
	else
		//  Write number of printers to output string:
		Sprintf( inString, L"%d", NumPrinters );


	//  Release the buffer:

	FreePrintersList();
	return 0;
}



int ConsoleShow( LPTSTR outString )
{
	if ( !QueryIsConsole( GetStdHandle( STD_OUTPUT_HANDLE ) ) )
		return 0;

	QPuts( L"\r                                                                      \r" );

	if ( outString )
		QPuts( outString );

	return 0;
}



int IsWin64()
{

#ifdef	_WIN64

	//  If app is compiled for 64-bit Windows, then duh....
	return 1;

#else	//	_WIN64

	//  App compiled for Win32; need to check whether it's actually running on Win64.

	typedef UINT ( __stdcall *GET64DIRPROC )( LPTSTR, UINT );
	HMODULE hKernel32 = NULL;
	GET64DIRPROC Get64Dir = NULL;
	wchar_t dirbuf[MAX_ARG_LEN] = L"";

	hKernel32 = GetModuleHandle( L"kernel32.dll" );

	if ( !hKernel32 )
		return 0;

	Get64Dir = (GET64DIRPROC) GetProcAddress( hKernel32, "GetSystemWow64DirectoryW" );

	if ( !Get64Dir )
		return 0;

	if ( (Get64Dir) ( dirbuf, MAX_ARG_LEN ) == 0 )
		return 0;

	if ( !iswalpha( dirbuf[0] ) )
		return 0;
	
	if ( dirbuf[1] != ':' )
		return 0;

	if ( dirbuf[2] != '\\' )
		return 0;
	else
		return 1;

#endif	//	_WIN64
}



int GetWindowsVersion( LPTSTR WindowsVer )
{
	if ( !WindowsVer )
		return -1;


	OSVERSIONINFOEXW VerInfo;
	typedef NTSTATUS ( WINAPI _RtlGetVersionProc ) ( OSVERSIONINFOEXW *VerInfo );
	HMODULE NtDllHandle						= GetModuleHandle( L"NtDll.dll" );
	_RtlGetVersionProc *RtlGetVersionProc	= NULL;


	memset( &VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo.dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	if ( NtDllHandle )
		RtlGetVersionProc = (_RtlGetVersionProc*) GetProcAddress( NtDllHandle, "RtlGetVersion" );

	if ( RtlGetVersionProc )
		RtlGetVersionProc( &VerInfo );

	if ( VerInfo.dwMajorVersion == 0 )
		GetVersionExW( (OSVERSIONINFOW*) &VerInfo );


	if ( ( VerInfo.dwMajorVersion == 5 ) && ( VerInfo.dwMinorVersion == 0 ) )
		wcscpy_s( WindowsVer, SHORT_BUF_LEN, L"Win2K" );
	else if ( ( VerInfo.dwMajorVersion == 5 ) && ( VerInfo.dwMinorVersion >= 1 ) && ( VerInfo.dwMinorVersion <= 2 ) )
		wcscpy_s( WindowsVer, SHORT_BUF_LEN, L"XP" );
	else if ( ( VerInfo.dwMajorVersion == 6 ) && ( VerInfo.dwMinorVersion == 0 ) )
		wcscpy_s( WindowsVer, SHORT_BUF_LEN, L"Vista" );
	else if ( ( VerInfo.dwMajorVersion == 6 ) && ( VerInfo.dwMinorVersion == 1 ) )
		wcscpy_s( WindowsVer, SHORT_BUF_LEN, L"Win7" );
	else if ( ( VerInfo.dwMajorVersion == 6 ) && ( VerInfo.dwMinorVersion >= 2 ) )
		wcscpy_s( WindowsVer, SHORT_BUF_LEN, L"Win8" );
	else if ( VerInfo.dwMajorVersion == 10 )
		wcscpy_s( WindowsVer, SHORT_BUF_LEN, L"Win10" );
	else
		wsprintf( WindowsVer, L"Win%d.%d", VerInfo.dwMajorVersion, VerInfo.dwMinorVersion );

	return 0;
}



int ParseFilenameTag( LPTSTR inString, size_t *index )
{
	if ( !inString )
		return 0;

	wchar_t shortbuf[SHORT_BUF_LEN] = L"";
	int i = 0;
	wchar_t ch = '\0';

	do {
		ch = inString[i];
		shortbuf[i++] = towupper( ch );
	} while ( ( ch != '>' ) && ( ch != '\0' ) && ( i < (SHORT_BUF_LEN - 1) ) );

	if ( ch != '>' )
		return 0;

	shortbuf[i - 1] = '\0';

	if ( !_wcsicmp( shortbuf, L"64" ) || !_wcsicmp( shortbuf, L"x64" ) ) {
		*index += i;
		return 1;
	}
	else if ( !_wcsicmp( shortbuf, L"win" ) || !_wcsicmp( shortbuf, L"ver" ) ) {
		*index += i;
		return 2;
	}

	return 0;
}



int InsertOSVersion( LPTSTR Filename )
{
	if ( !Filename )
		return -1;

	wchar_t WindowsVer[SHORT_BUF_LEN] = L"";
	GetWindowsVersion( WindowsVer );

	BOOL IsX64 = IsWin64();

	wchar_t Temp[MAX_ARG_LEN] = L"";
	size_t i = 0, j = 0;
	wchar_t ch = '\0';

	do {
		ch = Filename[i++];

		if ( ch == '<' ) {

			int tag = ParseFilenameTag( Filename + i, &i );

			switch ( tag ) {

				case 1:			// <x64>
					if ( IsX64 ) {
						wcscat_s( Temp, MAX_ARG_LEN, L"-x64" );
						j = wcslen( Temp );
					}

					break;

				case 2:			// <win>
					wcscat_s( Temp, MAX_ARG_LEN, WindowsVer );
					j = wcslen( Temp );
					break;

				default:
					Temp[j] = ch;
					if ( j < ( MAX_ARG_LEN - 1 ) )
						j++;

					Temp[j] = '\0';

			}
		}
		else {
			Temp[j] = ch;

			if ( j < ( MAX_ARG_LEN - 1 ) )
				j++;

			Temp[j] = '\0';
		}


	} while ( ch != '\0' );

	wcscpy_s( Filename, MAX_ARG_LEN, Temp );
	return 0;
}



int ReplaceChar( LPTSTR inString, wchar_t fromChar, wchar_t toChar )
{
	if ( !inString )
		return -1;

	int i = 0;
	wchar_t ch = '\0';

	do {
		if ( ( ch = inString[i] ) == fromChar )
			ch = toChar;

		inString[i++] = ch;
	
	} while ( ch != '\0' );

	return 0;
}



DLLExports int WINAPI defaultprint( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	if ( !numargs ) {

		DWORD bufsize = SHORT_BUF_LEN;

		if ( GetDefaultPrinter( inString, &bufsize ) )
			Printf( L"Default printer is \"%s\".\r\n", inString );
		else
			QPuts( L"No default printer.\r\n" );

		return 0;
	}
	else if ( !wcscmp( arg[0], L"/?" ) )
		return ShowCmdHelp( DefaultPrintHelpText, FALSE );
	else if ( !wcscmp( arg[0], L"/??" ) )
		return ShowCmdHelp( DefaultPrintHelpText, TRUE );

	if ( !SetDefaultPrinter( arg[0] ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error making \"%s\" default\r\n", arg[0] );
		return 2;
	}

	Printf( L"Default printer is now \"%s\".\r\n", arg[0] );
	return 0;
}



int MyDeletePrinter( LPTSTR PrinterName, BOOL KillIPPorts )
{
	if ( !PrinterName )
		return -1;

	HANDLE PrinterHandle = NULL;
	PRINTER_DEFAULTS Defaults = { NULL, NULL, PRINTER_ALL_ACCESS };

	PRINTER_INFO_2 * Info2;
	DWORD Attribs = 0;
	wchar_t PortName[SHORT_BUF_LEN] = L"";

	int rv;


	if ( !OpenPrinter( PrinterName, &PrinterHandle, &Defaults ) ) {
		rv = GetLastError();
		if ( !rv )
			rv = -1;

		return rv;
	}

	if ( !GetPrinterInfo( PrinterHandle, 2 ) ) {
		Info2 = (PRINTER_INFO_2 *) PrinterInfo;
		Attribs = Info2->Attributes;
		wcscpy_s( PortName, SHORT_BUF_LEN, Info2->pPortName );
	}
	else {
		Attribs = 0;
		wcscpy_s( PortName, SHORT_BUF_LEN, L"nul:" );
	}

	if ( Attribs & PRINTER_ATTRIBUTE_NETWORK ) {
		//  Network printer -- just remove the connection:
		
		ConsoleShow( L"Remove printer connection: " );
		
		ClosePrinter( PrinterHandle );

		if ( DeletePrinterConnection( PrinterName ) )
			rv = 0;
		else {
			rv = GetLastError();
			if ( !rv )
				rv = -1;
		}
	}
	else {
		//  Local printer -- waste it:

		ConsoleShow( L"Purge print queue: " );
		SetPrinter( PrinterHandle, 0, NULL, PRINTER_CONTROL_PURGE );

		ConsoleShow( L"Delete printer: " );

		if ( DeletePrinter( PrinterHandle ) )
			rv = 0;
		else {
			rv = GetLastError();
			if ( !rv )
				rv = -1;
		}

		ClosePrinter( PrinterHandle );

		if ( KillIPPorts ) {
			if ( IsIPPort( PortName ) ) {
				ConsoleShow( L"Remove IP port: " );
				RemoveIPPort( PortName );
			}
		}
	}

	ConsoleShow( NULL );
	FreePrinterInfo();
	return rv;
}



DLLExports int WINAPI delprinter( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	BYTE argtype[MAX_ARG_LEN];

	BOOL ExcludeName      = FALSE;
	BOOL IncludeName      = FALSE;
	BOOL ExactMatch       = FALSE;
	BOOL AnyPrinterSpecs  = FALSE;
	BOOL AnySpecsAtAll    = FALSE;
	BOOL KillIPPorts      = FALSE;
	BOOL NotReally        = FALSE;
	BOOL NoPrompts        = FALSE;
	BOOL DoIt             = FALSE;

	int PrintersFound = 0;
	int PrintersDeleted = 0;
	int rv = 0;
	int errorlevel = 0;

	wchar_t ErrorMessage[SHORT_BUF_LEN] = L"";

	for ( int i = 0; i < numargs; i++ ) {
		argtype[i] = 0;

		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( DelPrinterHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( DelPrinterHelpText, TRUE );
			else if ( !_wcsicmp( arg[i], L"/N" ) )
				NotReally = TRUE;
			else if ( !_wcsicmp( arg[i], L"/Y" ) )
				NoPrompts = TRUE;
			else if ( !_wcsicmp( arg[i], L"/K" ) )
				KillIPPorts = TRUE;
			else if ( ( arg[i][0] == '!' ) && ( arg[i][1] != '\0' ) ) {
				argtype[i] = 2;
				AnySpecsAtAll = TRUE;
			}
			else if ( arg[i][0] != '/' ) {
				argtype[i] = 1;
				AnySpecsAtAll = AnyPrinterSpecs = TRUE;
			}
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}

	if ( !AnySpecsAtAll ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing printer name\r\n" );
		return 1;
	}

	FreePrintersList();
	PRINTER_INFO_4 * Info4 = NULL;
	FreePortsList();


	//  Get list of printers:

	if ( GetPrintersList( 4, PRINTER_ENUM_LOCAL | PRINTER_ENUM_CONNECTIONS ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t get list of printers\r\n" );
		return 2;
	}

	Info4 = (PRINTER_INFO_4 *) PrintersList;


	//  Get list of ports:

	if ( GetPortsList() ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t get list of printer ports\r\n" );
		FreePrintersList();
		return 2;
	}


	//  Loop through list of printers:

	for ( unsigned int i = 0; i < NumPrinters; i++ ) {
		ExcludeName = FALSE;
		IncludeName = !AnyPrinterSpecs;
		ExactMatch  = FALSE;

		for ( int j = 0; j < numargs; j++ ) {
			if ( argtype[j] == 2 ) {
				if ( !WildcardComparison( arg[j]+1, Info4[i].pPrinterName, 0, 1 ) )
					ExcludeName = TRUE;
			}
			else if ( argtype[j] == 1 ) {
				if ( !_wcsicmp( arg[j], Info4[i].pPrinterName ) )
					ExactMatch = TRUE;
				else if ( !WildcardComparison( arg[j], Info4[i].pPrinterName, 0, 1 ) )
					IncludeName = TRUE;
			}
		}

		if ( !ExactMatch )
			if ( ExcludeName || !IncludeName )
				continue;

		DoIt = TRUE;

		if ( !NoPrompts && !ExactMatch ) {
			//  Prompt the user:

			Printf( L"Delete printer \"%s\"", Info4[i].pPrinterName );
			int k = QueryInputChar( L"", L"Y/N/A" );

			switch ( k ) {
				case 'A':
					NoPrompts = TRUE;
					// fall through
				case 'Y':
					DoIt = TRUE;
					break;
				default:
					DoIt = FALSE;
			}
		}

		PrintersFound++;

		if ( DoIt ) {
			if ( NotReally ) {
				Printf( L"Printer \"%s\" deleted.\r\n", Info4[i].pPrinterName );
				PrintersDeleted++;
			}
			else {
				rv = MyDeletePrinter( Info4[i].pPrinterName, KillIPPorts );

				if ( !rv ) {
					Printf( L"Printer \"%s\" deleted.\r\n", Info4[i].pPrinterName );
					PrintersDeleted++;
				}
				else {
					DisplayErrorHeader();
					GetErrorText( rv, ErrorMessage );
					Qprintf( ERRHANDLE, L"Can\u2019t delete printer \"%s\" (%s)\r\n", Info4[i].pPrinterName, ErrorMessage );
					errorlevel = 2;
				}
			}
		}

		QPuts( L"\r\n" );
	}


	//  Show summary report:

	if ( !PrintersFound )
		QPuts( L"No matching printers found.\r\n" );
	else
		Printf( L"%d matching %s found, %d %s.\r\n", PrintersFound, ( PrintersFound == 1 ) ? L"printer" : L"printers", PrintersDeleted, NotReally ? L"would be deleted" : L"deleted" );


	//  Clean up and exit:

	FreePrintersList();
	FreePortsList();
	return errorlevel;
}



int FreeJobs()
{
	if ( PrintJobs )
		free( PrintJobs );

	PrintJobs = NULL;
	return 0;
}



DLLExports int WINAPI listjobs( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_QUOTES_KLUDGE );


	wchar_t PrinterName[SHORT_BUF_LEN] = L"";

	DWORD bufsize = SHORT_BUF_LEN;
	HANDLE PrinterHandle = NULL;
	PRINTER_INFO_2 * Info2 = NULL;
	DWORD jobs;

	BYTE argtype[MAX_ARG_LEN];
	DWORD argval[MAX_ARG_LEN];
	BOOL ExcludeName  = FALSE;
	BOOL IncludeName  = FALSE;
	BOOL MatchJobID   = FALSE;
	BOOL AnyJobSpecs  = FALSE;
	BOOL AnyJobsFound = FALSE;

	wchar_t ErrorMessage[ SHORT_BUF_LEN ] = L"";
	int rv = 0;


	for ( int i = 0; i < numargs; i++ ) {
		argtype[i] = 0;
		argval[i]  = 0xffffffff;

		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( ListJobsHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( ListJobsHelpText, TRUE );
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				//  Get printer name:

				int j = 2;

				while ( iswalpha( arg[i][j] ) )
					j++;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;
				
				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					return 1;
				}

				wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[i]+j );
			}
			else if ( arg[i][0] != '/' ) {
				//  job wildspec or id: 
				
				if ( GetInt( arg[i] ) >= 0 ) {
					argtype[i] = 3;
					argval[i] = GetInt( arg[i] );
					AnyJobSpecs = TRUE;
				}
				else if ( ( arg[i][0] == '!' ) && ( arg[i][1] != '\0' ) )
					argtype[i] = 2;
				else {
					argtype[i] = 1;
					AnyJobSpecs = TRUE;
				}

			}
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}

	if ( PrinterName[0] == '\0' ) {

		if ( !GetDefaultPrinter( PrinterName, &bufsize ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default printer\r\n" );
			return 2;
		}
	}

	if ( !OpenPrinter( PrinterName, &PrinterHandle, NULL ) ) {
		rv = GetLastError();
		GetErrorText( rv, ErrorMessage );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening \"%s\" (%s)\r\n", PrinterName, ErrorMessage );
		return 2;
	}

	if ( GetPrinterInfo( PrinterHandle, 2 ) ) {
		rv = GetLastError();
		GetErrorText( rv, ErrorMessage );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error getting printer info for \"%s\" (%s)\r\n", PrinterName, ErrorMessage );
		ClosePrinter( PrinterHandle );
		return 2;
	}

	Info2 = (PRINTER_INFO_2 *) PrinterInfo;

	if ( Info2->cJobs == 0 ) {
		Printf( L"Printer \"%s\": No jobs.\r\n", PrinterName );
		FreePrinterInfo();
		ClosePrinter( PrinterHandle );
		return 0;
	}


	EnumJobs( PrinterHandle, 0, Info2->cJobs, 1, NULL, 0, &bufsize, &jobs );
	PrintJobs = (JOB_INFO_1 *) malloc( bufsize );

	if ( !EnumJobs( PrinterHandle, 0, Info2->cJobs, 1, (LPBYTE) PrintJobs, bufsize, &bufsize, &jobs ) ) {
		int rv = GetLastError();
		GetErrorText( rv, ErrorMessage );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error getting jobs list for \"%s\" (%s)\r\n", PrinterName, ErrorMessage );

		FreePrinterInfo();
		FreeJobs();
		ClosePrinter( PrinterHandle );
		return 2;
	}

	AnyJobsFound = FALSE;

	for ( unsigned int i = 0; i < Info2->cJobs; i++ ) {
		ExcludeName = FALSE;
		IncludeName = !AnyJobSpecs;
		MatchJobID  = FALSE;

		for ( int j = 0; j < numargs; j++ ) {
			if ( argtype[j] == 3 ) {
				if ( argval[j] == PrintJobs[i].JobId )
					MatchJobID = TRUE;
			}
			else if ( argtype[j] == 2 ) {
				if ( !WildcardComparison( arg[j]+1, PrintJobs[i].pDocument, 0, 1 ) )
					ExcludeName = TRUE;
			}
			else if ( argtype[j] == 1 ) {
				if ( !WildcardComparison( arg[j], PrintJobs[i].pDocument, 0, 1 ) )
					IncludeName = TRUE;
			}
		}

		if ( MatchJobID || ( IncludeName && ( !ExcludeName ) ) ) {
			Printf( L"%4d: %s\r\n", PrintJobs[i].JobId, PrintJobs[i].pDocument ? PrintJobs[i].pDocument : L"" );
			AnyJobsFound = TRUE;
		}

	}

	if ( !AnyJobsFound )
		Printf( L"Printer \"%s\": No matching jobs found.\r\n", PrinterName );

	FreePrinterInfo();
	FreeJobs();
	ClosePrinter( PrinterHandle );
	return 0;
}



DLLExports int WINAPI killjobs( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_QUOTES_KLUDGE );


	wchar_t PrinterName[SHORT_BUF_LEN] = L"";

	DWORD bufsize = SHORT_BUF_LEN;
	HANDLE PrinterHandle = NULL;
	PRINTER_DEFAULTS Defaults = { NULL, NULL, PRINTER_ALL_ACCESS };
	PRINTER_INFO_2 * Info2 = NULL;
	DWORD jobs;

	BYTE argtype[MAX_ARG_LEN];
	DWORD argval[MAX_ARG_LEN];
	BOOL ExcludeName  = FALSE;
	BOOL IncludeName  = FALSE;
	BOOL MatchJobID   = FALSE;
	BOOL AnyJobSpecs  = FALSE;
	BOOL AnyJobsFound = FALSE;
	BOOL KillFirst    = TRUE;
	int rv = 0;
	int errorlevel = 0;


	for ( int i = 0; i < numargs; i++ ) {
		argtype[i] = 0;
		argval[i]  = 0xffffffff;

		if ( arglen[i] ) {
			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( KillJobsHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( KillJobsHelpText, TRUE );
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				//  Get printer name:

				int j = 2;

				while ( iswalpha( arg[i][j] ) )
					j++;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;
				
				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					return 1;
				}

				wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[i]+j );
			}
			else if ( arg[i][0] != '/' ) {
				//  job wildspec or id: 

				KillFirst = FALSE;
				
				if ( GetInt( arg[i] ) >= 0 ) {
					argtype[i] = 3;
					argval[i] = GetInt( arg[i] );
					AnyJobSpecs = TRUE;
				}
				else if ( ( arg[i][0] == '!' ) && ( arg[i][1] != '\0' ) )
					argtype[i] = 2;
				else {
					argtype[i] = 1;
					AnyJobSpecs = TRUE;
				}

			}
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}

	if ( PrinterName[0] == '\0' ) {

		if ( !GetDefaultPrinter( PrinterName, &bufsize ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default printer\r\n" );
			return 2;
		}
	}

	if ( !OpenPrinter( PrinterName, &PrinterHandle, &Defaults ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening \"%s\"\r\n", PrinterName );
		return 2;
	}

	if ( GetPrinterInfo( PrinterHandle, 2 ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error getting printer info for \"%s\"\r\n", PrinterName );
		ClosePrinter( PrinterHandle );
		return 2;
	}

	Info2 = (PRINTER_INFO_2 *) PrinterInfo;

	if ( Info2->cJobs == 0 ) {
		Printf( L"Printer \"%s\": No jobs.\r\n", PrinterName );
		FreePrinterInfo();
		ClosePrinter( PrinterHandle );
		return 0;
	}


	EnumJobs( PrinterHandle, 0, Info2->cJobs, 1, NULL, 0, &bufsize, &jobs );
	PrintJobs = (JOB_INFO_1 *) malloc( bufsize );

	if ( !EnumJobs( PrinterHandle, 0, Info2->cJobs, 1, (LPBYTE) PrintJobs, bufsize, &bufsize, &jobs ) ) {
		int rv = GetLastError();
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error %d getting jobs list for \"%s\"\r\n", rv, PrinterName );

		FreePrinterInfo();
		FreeJobs();
		ClosePrinter( PrinterHandle );
		return 2;
	}

	AnyJobsFound = FALSE;

	for ( unsigned int i = 0; i < Info2->cJobs; i++ ) {
		ExcludeName = FALSE;
		IncludeName = !AnyJobSpecs;
		MatchJobID  = FALSE;

		for ( int j = 0; j < numargs; j++ ) {
			if ( argtype[j] == 3 ) {
				if ( argval[j] == PrintJobs[i].JobId )
					MatchJobID = TRUE;
			}
			else if ( argtype[j] == 2 ) {
				if ( !WildcardComparison( arg[j]+1, PrintJobs[i].pDocument, 0, 1 ) )
					ExcludeName = TRUE;
			}
			else if ( argtype[j] == 1 ) {
				if ( !WildcardComparison( arg[j], PrintJobs[i].pDocument, 0, 1 ) )
					IncludeName = TRUE;
			}
		}

		if ( KillFirst || MatchJobID || ( IncludeName && ( !ExcludeName ) ) ) {
			// Printf( L"Delete job %d: %s\r\n", PrintJobs[i].JobId, PrintJobs[i].pDocument ? PrintJobs[i].pDocument : L"" );
			AnyJobsFound = TRUE;

			if ( !SetJob( PrinterHandle, PrintJobs[i].JobId, 0, 0, JOB_CONTROL_DELETE ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Can\u2019t delete job #%d\r\n", PrintJobs[i].JobId );
				errorlevel = 2;
			}
			else
				Printf( L"Deleted job #%d: \"%s\"\r\n", PrintJobs[i].JobId, PrintJobs[i].pDocument ? PrintJobs[i].pDocument : L"" );

			if ( KillFirst )
				break;
		}

	}

	if ( !AnyJobsFound )
		Printf( L"Printer \"%s\": No matching jobs found.\r\n", PrinterName );


	FreePrinterInfo();
	FreeJobs();
	ClosePrinter( PrinterHandle );
	return errorlevel;
}



DLLExports int WINAPI listports( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	int Type = 0;
	BYTE argtype[MAX_ARG_LEN];
	BOOL ExcludeName   = FALSE;
	BOOL IncludeName   = FALSE;
	BOOL AnyPortSpecs  = FALSE;
	BOOL AnyPortsFound = FALSE;

	int rv = 0;
	wchar_t ErrorMessage[SHORT_BUF_LEN] = L"";


	for ( int i = 0; i < numargs; i++ ) {
		argtype[i] = 0;

		if ( arglen[i] ) {

			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( ListPortsHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( ListPortsHelpText, TRUE );
			else if ( !_wcsicmp( arg[i], L"/I" ) )
				Type |= 1;
			else if ( !_wcsicmp( arg[i], L"/N" ) )
				Type |= 2;
			else if ( ( arg[i][0] == '!' ) && ( arg[i][1] != '\0' ) )
				argtype[i] = 2;
			else if ( arg[i][0] != '/' ) {
				argtype[i] = 1;
				AnyPortSpecs = TRUE;
			}
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}

	FreePortsList();

	if ( !Type )
		Type = 3;


	//  Get list of printer ports:

	if ( GetPortsList() ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t get list of printer ports\r\n" );
		return 0;
	}


	//  Dump list to standard output:

	AnyPortsFound = FALSE;

	for ( unsigned int i = 0; i < NumPorts; i++ ) {
		ExcludeName = FALSE;
		IncludeName = !AnyPortSpecs;

		if ( ( Type & 1 ) == 0 )
			if ( PortsList[i].fPortType & PORT_TYPE_NET_ATTACHED )
				continue;

		if ( ( Type & 2 ) == 0 )
			if ( ( PortsList[i].fPortType & PORT_TYPE_NET_ATTACHED ) == 0 )
				continue;

		for ( int j = 0; j < numargs; j++ ) {
			if ( argtype[j] == 2 ) {
				if ( !WildcardComparison( arg[j]+1, PortsList[i].pPortName, 0, 1 ) )
					ExcludeName = TRUE;
			}
			else if ( argtype[j] == 1 ) {
				if ( !WildcardComparison( arg[j], PortsList[i].pPortName, 0, 1 ) )
					IncludeName = TRUE;
			}
		}

		if ( ExcludeName || !IncludeName )
			continue;

		Printf( L"%s\r\n", PortsList[i].pPortName );
		AnyPortsFound = TRUE;
	}

	if ( !AnyPortsFound )
		QPuts( L"No matching ports found.\r\n" );

	//  Release the buffer:

	FreePortsList();
	return 0;
}



int GetPortAddress( LPTSTR PortName, LPTSTR ServerName, int Info, LPTSTR outString )
{
	if ( !PortName || !outString )
		return -1;

	// Printf( L"GetPortAddress( %s, %s, %d, %x )\r\n", PortName, ServerName, Info, outString );

	HANDLE hXcv = INVALID_HANDLE_VALUE;
	wchar_t shortbuf[SHORT_BUF_LEN] = L"";
	PRINTER_DEFAULTS Defaults = { NULL, NULL, SERVER_ACCESS_ADMINISTER };

	DWORD BytesNeeded = 0;
	DWORD XcvStatus = 0;
	BOOL rv = FALSE;

	// const wchar_t MonitorName[] = L"Standard TCP/IP Port";


	if ( !ServerName )
		Sprintf( shortbuf, L",XcvPort %s", PortName );
	else if ( ServerName[0] == '\0' )
		Sprintf( shortbuf, L",XcvPort %s", PortName );
	else
		Sprintf( shortbuf, L"%s\\,XcvPort %s", ServerName, PortName );

	// Printf( L"Open string is \"%s\".\r\n", shortbuf );

	if ( !OpenPrinter( shortbuf, &hXcv, &Defaults ) ) {
		// Printf( L"OpenPrinter() failed: %d\r\n", GetLastError() );
		return -1;
	}

	rv = XcvData( hXcv, ( Info == 1 ) ? L"HostAddress" : L"IPAddress", NULL, 0, (PBYTE) outString, sizeof( shortbuf ), &BytesNeeded, &XcvStatus );

	ClosePrinter( hXcv );

	if ( ( XcvStatus != NO_ERROR ) || ( rv == FALSE ) ) {
		// Printf( L"XcvData() returned %d, status %d.\r\n", rv, XcvStatus );
		return -1;
	}

	return 0;
}



DLLExports int WINAPI listprinters( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	int Type = 0;
	BYTE argtype[MAX_ARG_LEN];
	BOOL IPOnly           = FALSE;
	BOOL Verbose          = FALSE;
	BOOL ExcludeName      = FALSE;
	BOOL IncludeName      = FALSE;
	BOOL AnyPrinterSpecs  = FALSE;
	BOOL AnyPrintersFound = FALSE;

	wchar_t PortAddress[SHORT_BUF_LEN] = L"";
	int Count = 0;

	int rv = 0;
	wchar_t ErrorMessage[SHORT_BUF_LEN] = L"";


	for ( int i = 0; i < numargs; i++ ) {
		argtype[i] = 0;

		if ( arglen[i] ) {

			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( ListPrintersHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( ListPrintersHelpText, TRUE );
			else if ( !_wcsicmp( arg[i], L"/L" ) )
				Type |= PRINTER_ENUM_LOCAL;
			else if ( !_wcsicmp( arg[i], L"/N" ) )
				Type |= PRINTER_ENUM_CONNECTIONS;
			else if ( !_wcsicmp( arg[i], L"/I" ) )
				IPOnly = TRUE;
			else if ( !_wcsicmp( arg[i], L"/V" ) )
				Verbose = TRUE;
			else if ( ( arg[i][0] == '!' ) && ( arg[i][1] != '\0' ) )
				argtype[i] = 2;
			else if ( arg[i][0] != '/' ) {
				argtype[i] = 1;
				AnyPrinterSpecs = TRUE;
			}
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}


	FreePrintersList();
	PRINTER_INFO_2 * Info2 = NULL;

	if ( IPOnly )
		Type = PRINTER_ENUM_LOCAL;

	if ( !Type )
		Type = PRINTER_ENUM_LOCAL | PRINTER_ENUM_CONNECTIONS;


	//  Get list of printers:

	if ( GetPrintersList( 2, Type ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t get list of printers\r\n" );
		return 2;
	}

	if ( IPOnly | Verbose )
		if ( GetPortsList() ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t get list of printer ports\r\n" );
			FreePrintersList();
			return 2;
		}


	Info2 = (PRINTER_INFO_2*) PrintersList;


	//  Dump list to stdout:

	AnyPrintersFound = FALSE;

	for ( unsigned int i = 0; i < NumPrinters; i++ ) {
		ExcludeName = FALSE;
		IncludeName = !AnyPrinterSpecs;

		for ( int j = 0; j < numargs; j++ ) {
			if ( argtype[j] == 2 ) {
				if ( !WildcardComparison( arg[j]+1, Info2[i].pPrinterName, 0, 1 ) )
					ExcludeName = TRUE;
			}
			else if ( argtype[j] == 1 ) {
				if ( !WildcardComparison( arg[j], Info2[i].pPrinterName, 0, 1 ) )
					IncludeName = TRUE;
			}
		}

		if ( ExcludeName || !IncludeName )
			continue;

		if ( IPOnly )
			if ( !IsIPPort( Info2[i].pPortName ) )
				continue;

		if ( Verbose ) {

			if ( Info2[i].Attributes & PRINTER_ATTRIBUTE_LOCAL ) {

				if ( IsIPPort( Info2[i].pPortName ) ) {

					GetPortAddress( Info2[i].pPortName, NULL, 0, PortAddress );

					if ( PortAddress[0] == '\0' )
						GetPortAddress( Info2[i].pPortName, NULL, 1, PortAddress );

					Printf( L"%-25.25s  ", PortAddress );

				}
				else
					Printf( L"%-25.25s  ", Info2[i].pPortName );

			}
			else {
				size_t j = 0;
				wchar_t ch = '\0';

				while ( Info2[i].pServerName[j] != '\0' ) {
					ch = towupper( Info2[i].pServerName[j] );
					PortAddress[j++] = ch;
					PortAddress[j] = '\0';
				}

				Printf( L"%-25.25s  ", PortAddress );
			}

		}

		Printf( L"%s\r\n", Info2[i].pPrinterName );
		AnyPrintersFound = TRUE;
		Count++;
	}

	if ( !AnyPrintersFound )
		QPuts( L"No matching printers found.\r\n" );

	//  Release the buffer:

	FreePrintersList();
	FreePortsList();
	return 0;
}



#define DATA_MODE_NONE		0
#define DATA_MODE_SAVE		1
#define DATA_MODE_RESTORE	2

#define DATA_INFO_2			0x0001
#define DATA_INFO_7			0x0002
#define DATA_INFO_COLOR		0x0004
#define DATA_INFO_DATA		0x0008
#define DATA_INFO_SECURITY	0x0010
#define DATA_INFO_GLOBAL	0x0020
#define DATA_INFO_MINIMAL	0x0040
#define DATA_INFO_USER		0x0080
#define DATA_INFO_ALL		0x00ff
#define DATA_FORCE_NAME		0x1000
#define DATA_KEEP_NAME		0x2000
#define DATA_NAME_FLAGS		0x3000
#define DATA_RESOLVE_PORT	0x4000



DLLExports int WINAPI printerdata( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	wchar_t PrinterName[SHORT_BUF_LEN]	= L"";
	wchar_t Filename[MAX_ARG_LEN]		= L"";

	BOOL BeVerbose = FALSE, DoNothing = FALSE;

	int Mode = DATA_MODE_NONE;
	DWORD Info = 0;


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcscmp( arg[i] , L"/?" ) )
				return ShowCmdHelp( PrinterDataHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( PrinterDataHelpText, TRUE );
			else if ( !_wcsicmp( arg[i], L"/$" ) )
				BeVerbose = TRUE;
			else if ( !_wcsicmp( arg[i], L"/$$" ) )
				BeVerbose = DoNothing = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/S", 2 ) ) {
				int j = 2;

				while ( iswalpha( arg[i][j] ) )
					j++;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				wcscpy_s( Filename, MAX_ARG_LEN, arg[i] + j );
				Mode = DATA_MODE_SAVE;
			}
			else if ( !_wcsnicmp( arg[i], L"/R", 2 ) ) {
				int j = 2;

				while ( iswalpha( arg[i][j] ) )
					j++;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				wcscpy_s( Filename, MAX_ARG_LEN, arg[i] + j );
				Mode = DATA_MODE_RESTORE;
			}
			else if ( !wcscmp( arg[i], L"/2" ) )
				Info |= DATA_INFO_2;
			else if ( !wcscmp( arg[i], L"/7" ) )
				Info |= DATA_INFO_7;
			else if ( !_wcsicmp( arg[i], L"/C" ) )
				Info |= DATA_INFO_COLOR;
			else if ( !_wcsicmp( arg[i], L"/D" )  )
				Info |= DATA_INFO_DATA;
			else if ( !_wcsicmp( arg[i], L"/Y" ) )
				Info |= DATA_INFO_SECURITY;
			else if ( !_wcsicmp( arg[i], L"/G" ) )
				Info |= DATA_INFO_GLOBAL;
			else if ( !_wcsicmp( arg[i], L"/M" ) )
				Info |= DATA_INFO_MINIMAL;
			else if ( !_wcsicmp( arg[i], L"/U" ) )
				Info |= DATA_INFO_USER;
			else if ( !_wcsicmp( arg[i], L"/F" ) )
				Info |= DATA_FORCE_NAME;
			else if ( !_wcsicmp( arg[i], L"/K" ) )
				Info |= DATA_KEEP_NAME;
			else if ( !_wcsicmp( arg[i], L"/P" ) )
				Info |= DATA_RESOLVE_PORT;
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( PrinterName[0] == '\0' )
				wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[i] );
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}


	if ( Filename[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing filename\r\n" );
		return 1;
	}

	if ( Mode == DATA_MODE_NONE ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Must specify either /S: or /R:\r\n" );
		return 1;
	}

	if ( PrinterName[0] == '\0' ) {
		DWORD bufsize = SHORT_BUF_LEN;

		if ( !GetDefaultPrinter( PrinterName, &bufsize ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default printer\r\n" );
			return 2;
		}
	}

	if ( ( Info & DATA_NAME_FLAGS ) == DATA_NAME_FLAGS )
		Info &= ~DATA_FORCE_NAME;

	if ( Mode == DATA_MODE_SAVE )
		Info &= DATA_INFO_ALL;


	if ( BeVerbose ) {
		Printf( L"PrinterName: \"%s\"\n", PrinterName );
		Printf( L"Filename:    \"%s\"\n", Filename );
		Printf( L"Info:        0x%04x\n\n", Info );
	}

	if ( DoNothing )
		return 0;


	if ( ( Mode == DATA_MODE_RESTORE ) && !QueryIsFile( Filename ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"File not found: \"%s\"\r\n", Filename );
		return 2;
	}


	wchar_t CmdStr[2][4] = { L"/Ss", L"/Sr" };
	wchar_t OpStr[2][10] = { L"saving", L"restoring" };
	wchar_t DirStr[2][5] = { L"to", L"from" };


	typedef int ( __stdcall *PRINTUIPROC )( HWND, HINSTANCE, LPTSTR, int );
	HMODULE hPrintUI = NULL;
	PRINTUIPROC PrintUIEntry = NULL;
	wchar_t commandbuf[MAX_ARG_LEN] = L"";
	int rv = 0;
	wchar_t ErrorText[SHORT_BUF_LEN] = L"";


	hPrintUI = LoadLibrary( L"printui.dll" );

	if ( !hPrintUI ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t load PRINTUI.DLL\r\n" );
		return 2;
	}

	PrintUIEntry = (PRINTUIPROC) GetProcAddress( hPrintUI, "PrintUIEntryW" );

	if ( !PrintUIEntry ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t get address for PrintUIEntryW\r\n" );
		FreeLibrary( hPrintUI );
		return 2;
	}

	Sprintf( commandbuf, L" %s /q /n \"%s\" /a \"%s\"", CmdStr[Mode - 1], PrinterName, Filename );

	if ( Info & DATA_INFO_2 )
		wcscat_s( commandbuf, MAX_ARG_LEN, L" 2" );

	if ( Info & DATA_INFO_7 )
		wcscat_s( commandbuf, MAX_ARG_LEN, L" 7" );

	if ( Info & DATA_INFO_COLOR )
		wcscat_s( commandbuf, MAX_ARG_LEN, L" c" );

	if ( Info & DATA_INFO_DATA )
		wcscat_s( commandbuf, MAX_ARG_LEN, L" d" );

	if ( Info & DATA_INFO_SECURITY )
		wcscat_s( commandbuf, MAX_ARG_LEN, L" s" );

	if ( Info & DATA_INFO_GLOBAL )
		wcscat_s( commandbuf, MAX_ARG_LEN, L" g" );

	if ( Info & DATA_INFO_MINIMAL )
		wcscat_s( commandbuf, MAX_ARG_LEN, L" m" );

	if ( Info & DATA_INFO_USER )
		wcscat_s( commandbuf, MAX_ARG_LEN, L" u" );

	if ( Info & DATA_FORCE_NAME )
		wcscat_s( commandbuf, MAX_ARG_LEN, L" f" );

	if ( Info & DATA_KEEP_NAME )
		wcscat_s( commandbuf, MAX_ARG_LEN, L" r" );

	if ( Info & DATA_RESOLVE_PORT )
		wcscat_s( commandbuf, MAX_ARG_LEN, L" p" );

	rv = (PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );

	FreeLibrary( hPrintUI );

	if ( rv ) {
		DisplayErrorHeader();
		GetErrorText( rv, ErrorText );
		Qprintf( ERRHANDLE, L"Error %s \"%s\" data %s \"%s\"\r\n(%s)\r\n", OpStr[Mode - 1], PrinterName, DirStr[Mode - 1], Filename, ErrorText );
		return 2;
	}

	return 0;
}



DLLExports int WINAPI printprefs( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t PrinterName[SHORT_BUF_LEN]		= L"";
	unsigned int NewColor					= 0;
	unsigned int NewDuplex					= 0;
	unsigned int NewOrient					= 0;
	unsigned int NewPaperSize				= 0;
	BOOL Changes							= FALSE;


	DEVMODE *Defaults						= NULL;

	int ChangeLevel							= 9;
	PRINTER_INFO_9 NewDefaults;

	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";
	HANDLE PrinterHandle					= NULL;
	DWORD BytesNeeded						= 0;
	BOOL ChangesMade						= FALSE;
	BOOL Display							= TRUE;
	int rv									= 2;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( PrintPrefsHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( PrintPrefsHelpText, TRUE );
			else if ( !_wcsnicmp( arg[i], L"/C", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				NewColor = GetInt( arg[i] + j );
				if ( NewColor < 0 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad value: \"%s\"\r\n", arg[i] + j );
					return 1;
				}
				else if ( ( NewColor < DMCOLOR_MONOCHROME ) || ( NewColor > DMCOLOR_COLOR ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"New color must be %u to %u: %u\r\n", DMCOLOR_MONOCHROME, DMCOLOR_COLOR, NewColor );
					return 1;
				}

				Changes = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/D", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				NewDuplex = GetInt( arg[i] + j );
				if ( NewDuplex < 0 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad value: \"%s\"\r\n", arg[i] + j );
					return 1;
				}
				else if ( ( NewDuplex < DMDUP_SIMPLEX ) || ( NewDuplex > DMDUP_HORIZONTAL ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"New duplex must be %u to %u: %u\r\n", DMDUP_SIMPLEX, DMDUP_HORIZONTAL, NewDuplex );
					return 1;
				}

				Changes = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/O", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				NewOrient = GetInt( arg[i] + j );
				if ( NewOrient < 0 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad value: \"%s\"\r\n", arg[i] + j );
					return 1;
				}
				else if ( ( NewOrient < DMORIENT_PORTRAIT ) || ( NewOrient > DMORIENT_LANDSCAPE ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"New orientation must be %u to %u: %u\r\n", DMORIENT_PORTRAIT, DMORIENT_LANDSCAPE, NewOrient );
					return 1;
				}

				Changes = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				int j = 2;
				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				NewPaperSize = GetInt( arg[i] + j );
				if ( NewPaperSize < 0 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad value: \"%s\"\r\n", arg[i] + j );
					return 1;
				}
				else if ( ( NewPaperSize < DMPAPER_LETTER ) || ( NewPaperSize > DMPAPER_LAST ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"New paper size must be %u to %u: %u\r\n", DMPAPER_LETTER, DMPAPER_LAST, NewPaperSize );
					return 1;
				}

				Changes = TRUE;
			}
			else if ( arg[i][0] == '/' ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option \"%s\"\r\n", arg[i] );
				return 1;
			}
			else if ( PrinterName[0] == '\0' )
				wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[i] );
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}


	if ( PrinterName[0] == '\0' ) {
		DWORD MaxSize = SHORT_BUF_LEN;
		GetDefaultPrinter( PrinterName, &MaxSize );
	}

	if ( PrinterName[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"No default printer\r\n" );
		return 2;
	}




	if ( OpenPrinter( PrinterName, &PrinterHandle, NULL ) ) {

		BytesNeeded = DocumentProperties( NULL, PrinterHandle, PrinterName, NULL, NULL, 0 );

		if ( BytesNeeded ) {
			Defaults = (DEVMODE*) malloc( BytesNeeded );

			if ( Defaults ) {

				if ( DocumentProperties( NULL, PrinterHandle, PrinterName, Defaults, NULL, DM_OUT_BUFFER ) == IDOK ) {

					rv = 0;

					if ( Changes ) {

						Defaults->dmFields = 0;

						if ( NewColor )
							Defaults->dmColor = NewColor, Defaults->dmFields |= DM_COLOR;

						if ( NewDuplex )
							Defaults->dmDuplex = NewDuplex, Defaults->dmFields |= DM_DUPLEX;

						if ( NewOrient )
							Defaults->dmOrientation = NewOrient, Defaults->dmFields |= DM_ORIENTATION;

						if ( NewPaperSize )
							Defaults->dmPaperSize = NewPaperSize, Defaults->dmFields |= DM_PAPERSIZE;

						NewDefaults.pDevMode = Defaults;


						if ( SetPrinter( PrinterHandle, ChangeLevel, (BYTE*) &NewDefaults, 0 ) )
							ChangesMade = TRUE;
						else {
							GetErrorText( GetLastError(), ErrorText );
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Failed to set defaults for \"%s\":\r\n", PrinterName );
							Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
							Display = FALSE;
							rv = 2;
						}
					}


					if ( Display ) {

						if ( ChangesMade )
							Printf( L"\r\nNew defaults for printer \"%s\":\r\n", PrinterName );
						else
							Printf( L"\r\nDefaults for printer \"%s\":\r\n", PrinterName );

						Printf( L"  Color mode   /C:%u  ", Defaults->dmColor );

						switch ( Defaults->dmColor ) {

						case DMCOLOR_COLOR :
							Printf( L"(color)" );
							break;

						case DMCOLOR_MONOCHROME :
							Printf( L"(black and white)" );
							break;

						default :
							Printf( L"(unknown)" );
						}

						Printf( L"\r\n  Duplex       /D:%u  ", Defaults->dmDuplex );

						switch ( Defaults->dmDuplex ) {

						case DMDUP_SIMPLEX :
							Printf( L"(single-sided)" );
							break;

						case DMDUP_HORIZONTAL :
							Printf( L"(flip on short edge)" );
							break;

						case DMDUP_VERTICAL :
							Printf( L"(flip on long edge)" );
							break;

						default :
							Printf( L"(unknown)" );
						}

						Printf( L"\r\n  Orientation  /O:%u  ", Defaults->dmOrientation );

						switch ( Defaults->dmOrientation ) {

						case DMORIENT_PORTRAIT :
							Printf( L"(portrait)" );
							break;

						case DMORIENT_LANDSCAPE :
							Printf( L"(landscape)" );
							break;

						default :
							Printf( L"(unknown)" );
						}

						Printf( L"\r\n  Paper size   /P:%u  ", Defaults->dmPaperSize );

						switch( Defaults->dmPaperSize ) {

						case DMPAPER_LETTER :
							Printf( L"(letter)" );
							break;

						case DMPAPER_TABLOID :
							Printf( L"(tabloid)" );
							break;

						case DMPAPER_LEDGER :
							Printf( L"(ledger)" );
							break;

						case DMPAPER_LEGAL :
							Printf( L"(legal)" );
							break;

						case DMPAPER_EXECUTIVE :
							Printf( L"(executive)" );
							break;

						case DMPAPER_A4 :
							Printf( L"(A4)" );
							break;

						case DMPAPER_A5 :
							Printf( L"(A5)" );
							break;

						case DMPAPER_ENV_9 :
							Printf( L"(#9 envelope)" );
							break;

						case DMPAPER_ENV_10 :
							Printf( L"(#10 envelope)" );
							break;

						default :
							Printf( L"(unknown)" );
						}

						Printf( L"\r\n\r\n" );
					}
				}
				else {
					GetErrorText( GetLastError(), ErrorText );
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Error getting DEVMODE structure\r\n" );
					Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
				}

				free( Defaults );
			}
			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Can\u2019t allocate DEVMODE structure\r\n" );
			}
		}
		else {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Can\u2019t get DEVMODE structure size\r\n" );
		}

		ClosePrinter( PrinterHandle );
	}
	else {
		GetErrorText( GetLastError(), ErrorText );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening printer \"%s\":\r\n", PrinterName );
		Qprintf( ERRHANDLE, L"   %s\r\n", ErrorText );
	}




	return rv;
}



DLLExports int WINAPI setprinter( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	wchar_t PrinterName[SHORT_BUF_LEN] = L"";
	DWORD bufsize = SHORT_BUF_LEN;

	HANDLE PrinterHandle = INVALID_HANDLE_VALUE;
	PRINTER_DEFAULTS Defaults = { NULL, NULL, PRINTER_ACCESS_ADMINISTER };

	PRINTER_INFO_6 Info6 = { 0 };
	DWORD BytesNeeded = 0;

	DWORD Command = 0;
	DWORD StatusSet = 0, StatusClear = 0xffffffff;
	DWORD Status = 0;

	int rv = 0;
	wchar_t ErrorMessage[SHORT_BUF_LEN] = L"";
	

	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcscmp( arg[i] , L"/?" ) )
				return ShowCmdHelp( SetPrinterHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( SetPrinterHelpText, TRUE );

			else if ( !_wcsnicmp( arg[i], L"/PU", 3 ) )
				Command = PRINTER_CONTROL_PURGE;
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) )
				Command = PRINTER_CONTROL_PAUSE;
			else if ( !_wcsnicmp( arg[i], L"/R", 2 ) )
				Command = PRINTER_CONTROL_RESUME;
			else if ( !_wcsnicmp( arg[i], L"/OF", 3 ) ) {
				Command = PRINTER_CONTROL_SET_STATUS;
				StatusSet = PRINTER_STATUS_OFFLINE;
				StatusClear = 0xffffffff;
			}
			else if ( !_wcsnicmp( arg[i], L"/ON", 3 ) ) {
				Command = PRINTER_CONTROL_SET_STATUS;
				StatusSet = 0;
				StatusClear = 0xffffffff ^ PRINTER_STATUS_OFFLINE;
			}
			else if ( ( arg[i][0] != '/' ) && ( PrinterName[0] == '\0' ) )
				wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[i] );

			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}


	if ( PrinterName[0] == '\0' ) {
		if ( !GetDefaultPrinter( PrinterName, &bufsize ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default printer\r\n" );
			return 2;
		}
	}

	if ( !Command ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing command\r\n" );
		return 1;
	}


	if ( !OpenPrinter( PrinterName, &PrinterHandle, &Defaults ) ) {
		rv = GetLastError();
		GetErrorText( rv, ErrorMessage );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening printer (%s)\r\n", ErrorMessage );
		return 2;
	}

	if ( Command == PRINTER_CONTROL_SET_STATUS ) {
		GetPrinter( PrinterHandle, 6, (LPBYTE) &Info6, sizeof( Info6 ), &BytesNeeded );
		Status = ( Info6.dwStatus & StatusClear ) | StatusSet;
	}

	if ( !SetPrinter( PrinterHandle, 0, ( Command == PRINTER_CONTROL_SET_STATUS ) ? (LPBYTE) Status : NULL, Command ) ) {
		rv = GetLastError();
		GetErrorText( rv, ErrorMessage );
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error on SetPrinter() (%s)\r\n", ErrorMessage );
		return 2;
	}


	ClosePrinter( PrinterHandle );

	Printf( L"Printer \"%s\" ", PrinterName );

	switch ( Command ) {

		case PRINTER_CONTROL_SET_STATUS:
			if ( StatusSet == PRINTER_STATUS_OFFLINE )
				QPuts( L"offline" );
			else if ( StatusClear == ( 0xffffffff ^ PRINTER_STATUS_OFFLINE ) )
				QPuts( L"online" );
			break;

		case PRINTER_CONTROL_PAUSE:
			QPuts( L"paused" );
			break;

		case PRINTER_CONTROL_RESUME:
			QPuts( L"resumed" );
			break;

		case PRINTER_CONTROL_PURGE:
			QPuts( L"purged" );
			break;
	}

	QPuts( L".\r\n" );

	return 0;
}



BOOL PrinterExists( LPTSTR PrinterName )
{
	HANDLE PrinterHandle = NULL;
	DWORD rv = 0;

	if ( OpenPrinter( PrinterName, &PrinterHandle, NULL ) ) {
		ClosePrinter( PrinterHandle );
		return TRUE;
	}

	rv = ( GetLastError() );

	if ( rv == 1801 ) 
		return FALSE;
	
	return TRUE;
}



DLLExports int WINAPI testpage( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	wchar_t PrinterName[SHORT_BUF_LEN] = L"";
	DWORD bufsize = SHORT_BUF_LEN;

	wchar_t ErrorMessage[SHORT_BUF_LEN] = L"";


	if ( arglen[0] ) {

		if ( !wcscmp( arg[0], L"/?" ) )
			return ShowCmdHelp( TestPageHelpText, FALSE );
		else if ( !wcscmp( arg[0], L"/??" ) )
			return ShowCmdHelp( TestPageHelpText, TRUE );
		else if ( arg[0][0] == '/' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown option \"%s\"\r\n", arg[0] );
			return 1;
		}
		else
			wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[0] );

	}

	if ( PrinterName[0] == '\0' ) {
		if ( !GetDefaultPrinter( PrinterName, &bufsize ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default printer\r\n" );
			return 2;
		}
	}


	if ( !PrinterExists( PrinterName ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Error opening printer \"%s\"\r\n", PrinterName );
		return 2;
	}


	typedef int ( __stdcall *PRINTUIPROC )( HWND, HINSTANCE, LPTSTR, int );
	HMODULE hPrintUI = NULL;
	PRINTUIPROC PrintUIEntry = NULL;
	wchar_t commandbuf[MAX_ARG_LEN] = L"";
	int rv = 0;

	hPrintUI = LoadLibrary( L"printui.dll" );

	if ( !hPrintUI ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t load PRINTUI.DLL\r\n" );
		return 2;
	}

	PrintUIEntry = (PRINTUIPROC) GetProcAddress( hPrintUI, "PrintUIEntryW" );

	if ( !PrintUIEntry ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t get address for PrintUIEntryW\r\n" );
		FreeLibrary( hPrintUI );
		return 2;
	}

	Sprintf( commandbuf, L" /k /q /n \"%s\"", PrinterName );
	rv = (PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );

	FreeLibrary( hPrintUI );

	if ( rv ) {
		DisplayErrorHeader();
		GetErrorText( rv, ErrorMessage );
		Qprintf( ERRHANDLE, L"Error sending test page to \"%s\" (%s)\r\n", PrinterName, ErrorMessage );
		return 2;
	}

	Printf( L"Test page sent to \"%s\".\r\n", PrinterName );
	return 0;
}



DLLExports int WINAPI addipport( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_QUOTES_KLUDGE );


	wchar_t PortName[SHORT_BUF_LEN]     = L"";
	wchar_t HostAddress[SHORT_BUF_LEN]  = L"";
	wchar_t LPRQueueName[SHORT_BUF_LEN] = L"";

	WORD TCPPort = 9100;

	BOOL TCPPortSpecified     = FALSE;
	BOOL HostNameIsDottedQuad = FALSE;
	BOOL LPRPrinterPort       = FALSE;
	BOOL BeVerbose            = FALSE;
	BOOL DoNothing            = FALSE;

	wchar_t ErrorMessage[SHORT_BUF_LEN] = L"";
	int rv = 0;


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( AddIPPortHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( AddIPPortHelpText, TRUE );
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				//  Get port name:

				int j = 2;

				while ( iswalpha( arg[i][j] ) )
					j++;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;
				
				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					return 1;
				}

				wcscpy_s( PortName, SHORT_BUF_LEN, arg[i]+j );
			}
			else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
				//  Get port number:

				int j = 2;

				while ( iswalpha( arg[i][j] ) )
					j++;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( j < 0 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					return 1;
				}

				long int p = GetInt( arg[i] + j );

				if ( ( p < 0 ) || ( p > 0xffff ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE,L"Invalid TCP port value: \"%s\"\r\n", arg[i] );
					return 1;
				}

				TCPPort = (WORD) p;
				TCPPortSpecified = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/Q", 2 ) ) {
				//  Get queue name:

				int j = 2;

				while ( iswalpha( arg[i][j] ) )
					j++;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					return 1;
				}

				wcscpy_s( LPRQueueName, SHORT_BUF_LEN, arg[i]+j );
			}
			else if ( !_wcsicmp( arg[i], L"/$" ) )
				BeVerbose = TRUE;
			else if ( !_wcsicmp( arg[i], L"/$$" ) )
				BeVerbose = DoNothing = TRUE;
			else if ( ( arg[i][0] != '/' ) && ( HostAddress[0] == '\0' ) )
				wcscpy_s( HostAddress, SHORT_BUF_LEN, arg[i] );

			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}


	if ( HostAddress[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing address\r\n" );
		return 1;
	}


	//  Check for a TCP port number in HostAddress:

	int k = -1;
	for ( int j = 0; HostAddress[j] != '\0'; j++ )
		if ( HostAddress[j] == ':' )
			k = j;

	if ( k > 0 )
		if ( iswdigit( HostAddress[k + 1] ) ) {
			long int p = GetInt( HostAddress + k + 1 );

			if ( ( p > 0 ) && ( p <= 0xffff ) ) {
				TCPPort = (WORD) p;
				TCPPortSpecified = TRUE;
				HostAddress[k] = '\0';
			}
		}

	HostNameIsDottedQuad = LooksLikeAnIP( HostAddress );
	LPRPrinterPort = ( ( LPRQueueName[0] != '\0' ) || ( TCPPortSpecified && ( TCPPort == 515 ) ) );

	if ( !TCPPortSpecified )
		TCPPort = LPRPrinterPort ? 515 : 9100;

	if ( LPRPrinterPort && ( LPRQueueName[0] == '\0' ) )
		wcscpy_s( LPRQueueName, SHORT_BUF_LEN, DefaultQueueName );

	if ( PortName[0] == '\0' )
		GeneratePortName( HostAddress, PortName );


	if ( wcslen( PortName ) >= MAX_PORTNAME_LEN ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Port name is %d characters max\r\n", MAX_PORTNAME_LEN - 1 );
		return 1;
	}

	if ( wcslen( LPRQueueName ) >= MAX_QUEUENAME_LEN ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"LPR queue name is %d characters max\r\n", MAX_QUEUENAME_LEN - 1 );
		return 1;
	}


	if ( BeVerbose ) {
		Printf( L"Port name:    \"%s\"\r\n", PortName );
		Printf( L"Host address: \"%s\"\r\n", HostAddress );
		Printf( L"TCP port:     %d\r\n", TCPPort );
		Printf( L"Protocol:     %s\r\n", LPRPrinterPort ? L"LPR" : L"Raw TCP" );
		Printf( L"Queue name:   \"%s\"\r\n", LPRQueueName );
	}

	if ( DoNothing )
		return 0;

	
	rv = CreateIPPort( PortName, HostNameIsDottedQuad ? NULL : HostAddress, HostNameIsDottedQuad ? HostAddress : NULL, LPRQueueName, LPRPrinterPort ? PROTOCOL_LPR_TYPE : PROTOCOL_RAWTCP_TYPE, TCPPort );

	if ( rv ) {
		DisplayErrorHeader();
		GetErrorText( rv, ErrorMessage );
		Qprintf( ERRHANDLE, L"Error creating IP port \"%s\" (%s)\r\n", PortName, ErrorMessage );
		return 2;
	}

	Printf( L"Created port \"%s\" for %s \"%s\".\r\n", PortName, HostNameIsDottedQuad ? L"IP address" : L"hostname", HostAddress );
	return 0;
}



DLLExports int WINAPI f_addipport( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t PortName[SHORT_BUF_LEN]			= L"";
	wchar_t HostAddress[SHORT_BUF_LEN]		= L"";
	wchar_t QueueName[SHORT_BUF_LEN]		= L"";

	wchar_t *sep1							= NULL;
	int t									= 0;

	WORD TCPPort				= 9100;
	BOOL HostNameIsDottedQuad	= FALSE;
	BOOL LPRPrinterPort			= FALSE;


	ParseArgs( String, PARSE_BREAK_COMMAS );


	if ( arglen[0] )
		wcscpy_s( HostAddress, SHORT_BUF_LEN, arg[0] );
	else {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing address\r\n" );
		return -1;
	}

	if ( arglen[1] )
		wcscpy_s( QueueName, SHORT_BUF_LEN, arg[1] );

	if ( arglen[2] )
		wcscpy_s( PortName, SHORT_BUF_LEN, arg[2] );


	//		Check for a TCP port number in HostAddress :

	if ( HostAddress[0] == ':' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", HostAddress );
		return -1;
	}

	sep1 = wcschr( HostAddress, ':' );

	if ( sep1 ) {
		*sep1 = '\0';

		if ( iswdigit( sep1[1] ) ) {
			t = GetInt( sep1 + 1 );

			if ( ( t > 0 ) && ( t <= 0xffff ) )
				TCPPort = (WORD) t;
		}
	}


	//		Sanity check arguments :

	HostNameIsDottedQuad = LooksLikeAnIP( HostAddress );

	if ( QueueName[0] || ( TCPPort = 515 ) )
		LPRPrinterPort = TRUE;

	if ( LPRPrinterPort && ( QueueName[0] == '\0' ) )
		wcscpy_s( QueueName, SHORT_BUF_LEN, DefaultQueueName );

	if ( wcslen( PortName ) >= MAX_PORTNAME_LEN ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Port name is %d characters max\r\n", MAX_PORTNAME_LEN );
		return -1;
	}

	if ( wcslen( QueueName ) >= MAX_QUEUENAME_LEN ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Queue name is %d characters max\r\n", MAX_QUEUENAME_LEN );
		return -1;
	}


	//		Do it:

	if ( CreateIPPort( PortName, HostNameIsDottedQuad ? NULL : HostAddress, HostNameIsDottedQuad ? HostAddress : NULL, QueueName, LPRPrinterPort ? PROTOCOL_LPR_TYPE : PROTOCOL_RAWTCP_TYPE, TCPPort ) )
		wcscpy_s( String, MAX_ARG_LEN, ErrorString );
	else
		wcscpy_s( String, MAX_ARG_LEN, PortName );

	return 0;
}



int SetPrinterPrefs( LPTSTR PrinterName, int Color, int Duplex, int Orientation, int PaperSize, BOOL Complain )
{
	if ( !PrinterName )
		return -1;
	else if ( PrinterName[0] == '\0' )
		return -1;

	
	HANDLE PrinterHandle				= NULL;
	DEVMODE *Defaults					= NULL;
	int ChangeLevel						= 9;
	PRINTER_INFO_9 NewDefaults;
	DWORD BytesNeeded					= 0;
	wchar_t ErrorText[SHORT_BUF_LEN]	= L"";

	int rv								= 0;


	if ( OpenPrinter( PrinterName, &PrinterHandle, NULL ) ) {

		BytesNeeded = DocumentProperties( NULL, PrinterHandle, PrinterName, NULL, NULL, 0 );

		if ( BytesNeeded ) {
			Defaults = (DEVMODE*) malloc( BytesNeeded );

			if ( Defaults ) {

				if ( DocumentProperties( NULL, PrinterHandle, PrinterName, Defaults, NULL, DM_OUT_BUFFER ) == IDOK ) {

					Defaults->dmFields = 0;

					if ( Color )
						Defaults->dmColor = Color, Defaults->dmFields |= DM_COLOR;

					if ( Duplex )
						Defaults->dmDuplex = Duplex, Defaults->dmFields |= DM_DUPLEX;

					if ( Orientation )
						Defaults->dmOrientation = Orientation, Defaults->dmFields |= DM_ORIENTATION;

					if ( PaperSize )
						Defaults->dmPaperSize = PaperSize, Defaults->dmFields |= DM_PAPERSIZE;

					NewDefaults.pDevMode = Defaults;

					if ( SetPrinter( PrinterHandle, ChangeLevel, (BYTE*) &NewDefaults, 0 ) )
						rv = 0;
					else {
						//	SetPrinter() failed:

						rv = 5;

						if ( Complain ) {
							GetErrorText( GetLastError(), ErrorText );
							DisplayErrorHeader();
							Qprintf( ERRHANDLE, L"Failed to set defaults for \"%s\"\n", PrinterName );
							Qprintf( ERRHANDLE, L"    %s\n", ErrorText );
						}
					}
				}
				else {
					//	DocumentProperties() failed:

					rv = 4;

					if ( Complain ) {
						DisplayErrorHeader();
						Qprintf( ERRHANDLE, L"Error getting DEVMODE structure\n" );
					}
				}

				free( Defaults );
			}
			else {
				//	malloc() failed:

				rv = 3;

				if ( Complain ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Can\u2019t allocate DEVMODE structure\n" );
				}
			}
		}
		else {
			//	DocumentProperties() returned 0:

			rv = 2;

			if ( Complain ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Can\u2019t get DEVMODE structure size\n" );
			}
		}

		ClosePrinter( PrinterHandle );
	}
	else {
		//	error opening printer:

		rv = 1;

		if ( Complain ) {
			GetErrorText( GetLastError(), ErrorText );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Error opening printer \"%s\":\n", PrinterName );
			Qprintf( ERRHANDLE, L"    %s\n", ErrorText );
		}
	}


	return rv;
}



int IsNetPrinterName( LPTSTR inString )
{
	if ( !inString )
		return 0;

	if ( inString[0] != '\\' )
		return 0;

	if ( inString[1] != '\\' )
		return 0;

	if ( ( inString[2] == '\\' ) || ( inString[2] == '\0' ) )
		return 0;

	int i = 2;
	while ( ( inString[i] != '\\' ) && ( inString[i] != '\0' ) )
		i++;

	if ( inString[i] != '\\' )
		return 0;

	if ( ( inString[i + 1] == '\\' ) || ( inString[i] == '\0' ) )
		return 0;

	return 1;
}



int TweakPortName( LPTSTR PortName )
{
	if ( !PortName )
		return -1;
	else if ( PortName[0] == '\0' )
		return -1;


	BOOL knownport = FALSE;

	knownport |= ( ( towupper( PortName[0] ) == 'L' ) && ( towupper( PortName[1] ) == 'P' ) && ( towupper( PortName[2] ) == 'T' ) );
	knownport |= ( ( towupper( PortName[0] ) == 'C' ) && ( towupper( PortName[1] ) == 'O' ) && ( towupper( PortName[2] ) == 'M' ) );

	if ( !knownport )
		return 0;

	size_t i = 3;
	if ( !iswdigit( PortName[i] ) )
		return 0;

	while ( iswdigit( PortName[i] ) )
		i++;

	if ( ( PortName[i] == '\0' ) || ( ( PortName[i] == ':' ) && ( PortName[i + 1] == '\0' ) ) )
		for ( size_t j = 0; j < i; j++ )
			PortName[j] = towupper( PortName[j] );

	if ( PortName[i] != '\0' )
		return 0;

	if ( i < ( SHORT_BUF_LEN - 1 ) ) {
		PortName[i++] = ':';
		PortName[i] = '\0';
	}

	return 0;
}



DLLExports int WINAPI addprinter( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_QUOTES_KLUDGE );


	wchar_t PrinterName[SHORT_BUF_LEN]		= L"";
	wchar_t DriverName[SHORT_BUF_LEN]		= L"";
	wchar_t PortName[SHORT_BUF_LEN]			= L"";
	wchar_t HostAddress[SHORT_BUF_LEN]		= L"";
	wchar_t LPRQueueName[SHORT_BUF_LEN]		= L"";
	wchar_t InfFilename[MAX_ARG_LEN]		= L"";
	wchar_t PtrComment[SHORT_BUF_LEN]		= L"";
	wchar_t PtrLocation[SHORT_BUF_LEN]		= L"";
	wchar_t DataFilename[MAX_ARG_LEN]		= L"";
	wchar_t ErrorMessage[SHORT_BUF_LEN]		= L"";

	WORD TCPPort = 9100;

	BOOL TestPage				= FALSE;
	BOOL MakeDefault			= FALSE;
	BOOL AdditionalSettings		= FALSE;
	BOOL NoOverwrite			= FALSE;
	BOOL BeVerbose				= FALSE;
	BOOL DoNothing				= FALSE;
	BOOL TCPPortSpecified		= FALSE;
	BOOL HostNameIsDottedQuad	= FALSE;
	BOOL LPRPrinterPort			= FALSE;

	int PrefColor				= 0;
	int PrefDuplex				= 0;
	int PrefOrient				= 0;
	int PrefPaper				= 0;
	BOOL ChangePrefs			= FALSE;

	int rv						= 0;


	for ( int i = 0; i < numargs; i++ ) {
		if ( arglen[i] ) {

			if ( !wcscmp( arg[i], L"/?" ) )
				return ShowCmdHelp( AddPrinterHelpText, FALSE );
			else if ( !wcscmp( arg[i], L"/??" ) )
				return ShowCmdHelp( AddPrinterHelpText, TRUE );
			else if ( !_wcsicmp( arg[i], L"/T" ) )
				TestPage = TRUE;
			else if ( !_wcsicmp( arg[i], L"/D" ) )
				MakeDefault = TRUE;
			else if ( !_wcsicmp( arg[i], L"/O" ) )
				NoOverwrite = TRUE;
			else if ( !_wcsicmp( arg[i], L"/$" ) )
				BeVerbose = TRUE;
			else if ( !_wcsicmp( arg[i], L"/$$" ) )
				BeVerbose = DoNothing = TRUE;
			else if ( !_wcsnicmp( arg[i], L"/A", 2 ) ) {
				//  Get host address:

				int j = 2;

				while ( iswalpha( arg[i][j] ) )
					j++;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;
				
				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					return 1;
				}

				wcscpy_s( HostAddress, SHORT_BUF_LEN, arg[i]+j );

			}
			else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
				//  Get port name:

				int j = 2;

				while ( iswalpha( arg[i][j] ) )
					j++;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;
				
				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					return 1;
				}

				wcscpy_s( PortName, SHORT_BUF_LEN, arg[i]+j );
			}
			else if ( !_wcsnicmp( arg[i], L"/Q", 2 ) ) {
				//  Get LPR queue name:

				int j = 2;

				while ( iswalpha( arg[i][j] ) )
					j++;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;;

				if ( arg[i][j] == '\0' )
					wcscpy_s( LPRQueueName, SHORT_BUF_LEN, DefaultQueueName );
				else
					wcscpy_s( LPRQueueName, SHORT_BUF_LEN, arg[i]+j );
			}
			else if ( !_wcsnicmp( arg[i], L"/I", 2 ) ) {
				//  Get INF filename:

				int j = 2;

				while ( iswalpha( arg[i][j] ) )
					j++;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;
				
				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					return 1;
				}

				wcscpy_s( InfFilename, MAX_ARG_LEN, arg[i]+j );
			}
			else if ( !_wcsnicmp( arg[i], L"/C", 2 ) ) {
				//  Get printer comment:

				int j = 2;

				while ( iswalpha( arg[i][j] ) )
					j++;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;
				
				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					return 1;
				}

				wcscpy_s( PtrComment, SHORT_BUF_LEN, arg[i]+j );
				AdditionalSettings = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/L", 2 ) ) {
				//  Get printer location:

				int j = 2;

				while ( iswalpha( arg[i][j] ) )
					j++;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;
				
				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
					return 1;
				}

				wcscpy_s( PtrLocation, SHORT_BUF_LEN, arg[i]+j );
				AdditionalSettings = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/R", 2 ) ) {
				//  Get printer data filename:

				int j = 2;

				while ( iswalpha( arg[i][j] ) )
					j++;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				if ( arg[i][j] == '\0' ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Missing filename: \"%s\"\r\n", arg[i] );
					return 1;
				}

				wcscpy_s( DataFilename, MAX_ARG_LEN, arg[i] + j );
			}
			else if ( !_wcsnicmp( arg[i], L"/XC", 3 ) ) {
				//	Change default color setting:

				int j = 3;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				PrefColor = GetInt( arg[i] + j );

				if ( PrefColor < 0 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad value: \"%s\"\n", arg[i] + j );
					return 1;
				}
				else if ( ( PrefColor < DMCOLOR_MONOCHROME ) || ( PrefColor > DMCOLOR_COLOR ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Default color must be %u to %u: %u\n", DMCOLOR_MONOCHROME, DMCOLOR_COLOR, PrefColor );
					return 1;
				}

				ChangePrefs = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/XD", 3 ) ) {
				//	Change default duplex setting:

				int j = 3;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				PrefDuplex = GetInt( arg[i] + j );

				if ( PrefDuplex < 0 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad value: \"%s\"\n", arg[i] + j );
					return 1;
				}
				else if ( ( PrefDuplex < DMDUP_SIMPLEX ) || ( PrefDuplex > DMDUP_HORIZONTAL ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Default duplex must be %u to %u: %u\n", DMDUP_SIMPLEX, DMDUP_HORIZONTAL, PrefDuplex );
					return 1;
				}

				ChangePrefs = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/XO", 3 ) ) {
				//	Change default orientation setting:

				int j = 3;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				PrefOrient = GetInt( arg[i] + j );

				if ( PrefOrient < 0 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad value: \"%s\"\n", arg[i] + j );
					return 1;
				}
				else if ( ( PrefOrient < DMORIENT_PORTRAIT ) || ( PrefOrient > DMORIENT_LANDSCAPE ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Default orientation must be %u to %u: %u\n", DMORIENT_PORTRAIT, DMORIENT_LANDSCAPE, PrefOrient );
					return 1;
				}

				ChangePrefs = TRUE;
			}
			else if ( !_wcsnicmp( arg[i], L"/XP", 3 ) ) {
				//	Change default paper size setting:

				int j = 3;

				if ( IsColonOrEquals( arg[i][j] ) )
					j++;

				PrefPaper = GetInt( arg[i] + j );

				if ( PrefPaper < 0 ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Bad value: \"%s\"\n", arg[i] + j );
					return 1;
				}
				else if ( ( PrefPaper < DMPAPER_LETTER ) || ( PrefPaper > DMPAPER_LAST ) ) {
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Default paper size must be %u to %u: %u\n", DMPAPER_LETTER, DMPAPER_LAST, PrefPaper );
					return 1;
				}

				ChangePrefs = TRUE;
			}
			else if ( ( arg[i][0] != '/' ) && ( PrinterName[0] == '\0' ) )
				wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[i] );
			else if ( ( arg[i][0] != '/' ) && PrinterName[0] && ( DriverName[0] == '\0' ) )
				wcscpy_s( DriverName, SHORT_BUF_LEN, arg[i] );

			else {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
				return 1;
			}

		}
	}


	if ( PrinterName[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing printer name (/? for syntax help)\r\n" );
		return 1;
	}
	else if ( IsNetPrinterName( PrinterName ) ) {

		//  Connect to network printer:

		rv = 0;

		if ( ( DriverName[0] != '\0' ) || ( PortName[0] != '\0' ) || ( HostAddress[0] != '\0' ) || ( LPRQueueName[0] != '\0' ) || ( InfFilename[0] != '\0' ) || ( PtrComment[0] != '\0' ) || ( PtrLocation[0] != '\0' ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Don\u2019t combine network printer name with other options." );
			return 1;
		}

		if ( BeVerbose )
			Printf( L"Network printer name: \"%s\"\r\n", PrinterName );

		if ( DoNothing )
			return 0;

		if ( NoOverwrite ) {
			FreePrintersList();
			PRINTER_INFO_4 * Info4 = NULL;
			GetPrintersList( 4, PRINTER_ENUM_CONNECTIONS );
			Info4 = (PRINTER_INFO_4 *) PrintersList;

			rv = 0;

			if ( PrintersList )
				for ( unsigned int i = 0; i < NumPrinters; i++ )
					if ( !_wcsicmp( Info4[i].pPrinterName, PrinterName ) )
						rv = 1;

			FreePrintersList();

			if ( rv ) {
				Printf( L"Printer connection \"%s\" already exists; exiting.\r\n", PrinterName );
				return 0;
			}
		}


		ConsoleShow( L"Connect to printer: " );

		if ( !AddPrinterConnection( PrinterName ) ) {

			rv = GetLastError();
			if ( !rv )
				rv = -1;

			ConsoleShow( NULL );
			DisplayErrorHeader();
			GetErrorText( rv, ErrorMessage );
			Qprintf( ERRHANDLE, L"Failed to connect to printer \"%s\" (%s)\r\n", PrinterName, ErrorMessage );
			return 2;
		}

		if ( ChangePrefs )
			if ( SetPrinterPrefs( PrinterName, PrefColor, PrefDuplex, PrefOrient, PrefPaper, TRUE ) )
				rv = 2;

		if ( TestPage || MakeDefault ) {

			typedef int ( __stdcall *PRINTUIPROC )( HWND, HINSTANCE, LPTSTR, int );
			HMODULE hPrintUI = NULL;
			PRINTUIPROC PrintUIEntry = NULL;
			wchar_t commandbuf[MAX_ARG_LEN] = L"";


			hPrintUI = LoadLibrary( L"printui.dll" );

			if ( !hPrintUI ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Can\u2019t load PRINTUI.DLL\r\n" );
				return 2;
			}

			PrintUIEntry = (PRINTUIPROC) GetProcAddress( hPrintUI, "PrintUIEntryW" );

			if ( !PrintUIEntry ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Can\u2019t get address for PrintUIEntryW\r\n" );
				FreeLibrary( hPrintUI );
				return 2;
			}

			if ( TestPage ) {
				ConsoleShow( L"Send test page: " );
				
				Sprintf( commandbuf, L" /k /q /n \"%s\"", PrinterName );
				(PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );
			}

			if ( MakeDefault ) {
				ConsoleShow( L"Set default: " );
				SetDefaultPrinter( PrinterName );
			}
		}

		ConsoleShow( NULL );
		Printf( L"Connected to printer \"%s\".\r\n", PrinterName );
		return rv;
	}


	if ( DriverName[0] == '\0' ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing driver name\r\n" );
		return 1;
	}

	if ( ( PortName[0] == '\0' ) && ( HostAddress[0] == '\0' ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Must specify either /A:address or /P:port\r\n" );
		return 1;
	}
	else if ( ( PortName[0] != '\0' ) && ( HostAddress[0] != '\0' ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Specify only one of /A:address or /P:port\r\n" );
		return 1;
	}

	if ( ( HostAddress[0] == '\0' ) && ( LPRQueueName[0] != '\0' ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Use /Q:queue in conjunction with /A:address\r\n" );
		return 1;
	}


	if ( HostAddress[0] != '\0' ) {
		//  Check for a TCP port number in HostAddress:

		int k = -1;
		for ( int j = 0; HostAddress[j] != '\0'; j++ )
			if ( HostAddress[j] == ':' )
				k = j;

		if ( k > 0 )
			if ( iswdigit( HostAddress[k + 1] ) ) {
				long int p = GetInt( HostAddress + k + 1 );

				if ( ( p >= 0 ) && ( p <= 0xffff ) ) {
					TCPPort = (WORD) p;
					TCPPortSpecified = TRUE;
					HostAddress[k] = '\0';
				}
			}

		HostNameIsDottedQuad = LooksLikeAnIP( HostAddress );
		LPRPrinterPort = ( ( LPRQueueName[0] != '\0' ) || ( TCPPortSpecified && ( TCPPort == 515 ) ) );

		if ( !TCPPortSpecified )
			TCPPort = LPRPrinterPort ? 515 : 9100;

		if ( LPRPrinterPort && ( LPRQueueName[0] == '\0' ) )
			wcscpy_s( LPRQueueName, SHORT_BUF_LEN, DefaultQueueName );

	}



	if ( wcslen( PrinterName ) > 220 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Printer name is %d characters max\r\n", 220 );
		return 1;
	}

	if ( wcslen( PortName ) >= MAX_PORTNAME_LEN ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Port name is %d characters max\r\n", MAX_PORTNAME_LEN - 1 );
		return 1;
	}

	if ( wcslen( LPRQueueName ) >= MAX_QUEUENAME_LEN ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"LPR queue name is %d characters max\r\n", MAX_QUEUENAME_LEN - 1 );
		return 1;
	}

	if ( wcslen( PtrComment ) > 255 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Printer comment is %d characters max\r\n", 255 );
		return 1;
	}

	if ( wcslen( PtrLocation ) > 255 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Printer location is %d characters max\r\n", 255 );
		return 1;
	}


	if ( InfFilename[0] ) {
		ReplaceChar( InfFilename, '/', '\\' );
		InsertOSVersion( InfFilename );
	}

	TweakPortName( PortName );

	if ( DataFilename[0] ) {
		ReplaceChar( DataFilename, '/', '\\' );
		InsertOSVersion( DataFilename );
	}


	if ( BeVerbose ) {
		Printf( L"Printer name:  \"%s\"\r\n", PrinterName );
		Printf( L"Driver name:   \"%s\"\r\n", DriverName );
		Printf( L"Port name:     \"%s\"\r\n", PortName );
		Printf( L"Host address:  \"%s\"\r\n", HostAddress );
		Printf( L"TCP port:      %d\r\n", TCPPort );
		Printf( L"Protocol:      %s\r\n", LPRPrinterPort ? L"LPR" : L"Raw TCP" );
		Printf( L"Queue name:    \"%s\"\r\n", LPRQueueName );
		Printf( L"INF filename:  \"%s\"\r\n", InfFilename );
		Printf( L"Comment:       \"%s\"\r\n", PtrComment );
		Printf( L"Location:      \"%s\"\r\n", PtrLocation );
		Printf( L"Data filename: \"%s\"\r\n\n", DataFilename );
	}

	if ( DoNothing )
		return 0;

	if ( DataFilename[0] ) {
		if ( !QueryIsFile( DataFilename ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"File not found: \"%s\"\r\n", DataFilename );
			return 2;
		}
	}

	if ( NoOverwrite )
		if ( PrinterExists( PrinterName ) ) {
			Printf( L"Printer \"%s\" already exists; exiting.\r\n", PrinterName );
			return 0;
		}

	if ( HostAddress[0] != '\0' ) {
		rv = CreateIPPort( PortName, HostNameIsDottedQuad ? NULL : HostAddress, HostNameIsDottedQuad ? HostAddress : NULL, LPRQueueName, LPRPrinterPort ? PROTOCOL_LPR_TYPE : PROTOCOL_RAWTCP_TYPE, TCPPort );

		if ( rv ) {
			DisplayErrorHeader();
			GetErrorText( rv, ErrorMessage );
			Qprintf( ERRHANDLE, L"Error creating IP port \"%s\" (%s)\r\n", PortName, ErrorMessage );
			return 2;
		}
	}


	typedef int ( __stdcall *PRINTUIPROC )( HWND, HINSTANCE, LPTSTR, int );
	HMODULE hPrintUI = NULL;
	PRINTUIPROC PrintUIEntry = NULL;
	wchar_t commandbuf[MAX_ARG_LEN] = L"";


	hPrintUI = LoadLibrary( L"printui.dll" );

	if ( !hPrintUI ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t load PRINTUI.DLL\r\n" );
		return 2;
	}

	PrintUIEntry = (PRINTUIPROC) GetProcAddress( hPrintUI, "PrintUIEntryW" );

	if ( !PrintUIEntry ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Can\u2019t get address for PrintUIEntryW\r\n" );
		FreeLibrary( hPrintUI );
		return 2;
	}

	if ( !NoOverwrite ) {
		ConsoleShow( L"Delete old printer: " );

		Sprintf( commandbuf, L" /dl /n \"%s\" /q", PrinterName );
		// Printf( L"Command string: %s\r\n", commandbuf );
		(PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );
	}

    if ( InfFilename[0] != '\0' ) {

            ConsoleShow( L"Install driver: " );

            Sprintf( commandbuf, L" /ia /m \"%s\" /f \"%s\" /q", DriverName, InfFilename );
            // Printf( L"Command string: %s\r\n", commandbuf );
            rv = (PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );

            if ( rv ) {
                    ConsoleShow( NULL );
                    DisplayErrorHeader();
                    GetErrorText( rv, ErrorMessage );
                    Qprintf( ERRHANDLE, L"Error installing driver (%s)\r\n", ErrorMessage );
                    FreeLibrary( hPrintUI );
                    return 2;
            }
    }

	ConsoleShow( L"Create new printer: " );

	Sprintf( commandbuf, L" /if /b \"%s\" %s%s%s /q /r \"%s\" /m \"%s\" /z /u", PrinterName, ( ( InfFilename[0] == '\0' ) ? L"" : L"/f \"" ), ( ( InfFilename[0] == '\0' ) ? L"" : InfFilename ), ( ( InfFilename[0] == '\0' ) ? L"" : L"\" " ), PortName, DriverName );
	// Printf( L"Command string: %s\r\n", commandbuf );

	rv = (PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );

	if ( !rv ) {

		//  Created the printer successfully!

		/*
		if ( DataFilename[0] || AdditionalSettings || TestPage || MakeDefault ) {
			ConsoleShow( L"Wait a bit:" );
			Sleep( 500 );
		}
		*/

		if ( DataFilename[0] ) {
			ConsoleShow( L"Restore data file: " );

			Sprintf( commandbuf, L" /Sr /q /n \"%s\" /a \"%s\" d g r", PrinterName, DataFilename );
			// Printf( L"Command string: %s\r\n", commandbuf );

			int rv2 = (PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );

			if ( rv2 ) {
				ConsoleShow( NULL );
				DisplayErrorHeader();
				GetErrorText( rv2, ErrorMessage );
				Qprintf( ERRHANDLE, L"Error loading data file \"%s\":\r\n", DataFilename );
				Qprintf( ERRHANDLE, L"%s\r\n", ErrorMessage );
				CrLf();
			}
		}

		if ( AdditionalSettings ) {
			ConsoleShow( L"Additional settings: " );

			Sprintf( commandbuf, L" /Xs /q /n \"%s\"", PrinterName );

			if ( PtrComment[0] != '\0' ) {
				wcscat_s( commandbuf, MAX_ARG_LEN, L" comment \"" );
				wcscat_s( commandbuf, MAX_ARG_LEN, PtrComment );
				wcscat_s( commandbuf, MAX_ARG_LEN, L"\"" );
			}

			if ( PtrLocation[0] != '\0' ) {
				wcscat_s( commandbuf, MAX_ARG_LEN, L" location \"" );
				wcscat_s( commandbuf, MAX_ARG_LEN, PtrLocation );
				wcscat_s( commandbuf, MAX_ARG_LEN, L"\"" );
			}
			
			// Printf( L"Command string: %s\r\n", commandbuf );
			(PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );
		}

		if ( TestPage ) {
			ConsoleShow( L"Send test page: " );

			Sprintf( commandbuf, L" /k /q /n \"%s\"", PrinterName );
			(PrintUIEntry) ( GetForegroundWindow(), NULL, commandbuf, 0 );
		}

		if ( MakeDefault ) {
			ConsoleShow( L"Set default: " );
			SetDefaultPrinter( PrinterName );
		}

	}

	FreeLibrary( hPrintUI );

	ConsoleShow( NULL );

	if ( rv ) {
		DisplayErrorHeader();
		GetErrorText( rv, ErrorMessage );
		Qprintf( ERRHANDLE, L"Failed to create printer (%s)\r\n", ErrorMessage );
		return 2;
	}
	else {
		rv = 0;
		Printf( L"Created printer \"%s\".\r\n", PrinterName );

		if ( ChangePrefs )
			if ( SetPrinterPrefs( PrinterName, PrefColor, PrefDuplex, PrefOrient, PrefPaper, TRUE ) )
				rv = 2;
	}

	return rv;
}



DLLExports int WINAPI f_numports( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_COMMAS );


	wchar_t PortName[SHORT_BUF_LEN] = L"*";
	int count = 0, Type = 0;

	if ( arglen[0] ) {
		for ( int i = 0; arg[0][i] != '\0'; i++ ) {

			switch ( towupper( arg[0][i] ) ) {

				case 'I' :			// include ip ports
					Type |= 1;
					break;

				case 'N' :			// include non-ip ports
					Type |= 2;
					break;

				default :
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Invalid type \"%s\"\r\n", arg[0] );
					return -1;

			}
		}
	}

	if ( !Type )			// if no port type specified,
		Type = 3;			// include both

	if ( arglen[1] )
		wcscpy_s( PortName, SHORT_BUF_LEN, arg[1] );

	FreePortsList();

	if ( GetPortsList() ) {
		wcscpy_s( inString, SHORT_BUF_LEN, ErrorString );
		return 0;
	}

	if ( !NumPorts ) {
		wcscpy_s( inString, SHORT_BUF_LEN, L"0" );
		FreePortsList();
		return 0;
	}

	for ( unsigned int i = 0; i < NumPorts; i++ )
		if ( !WildcardComparison( PortName, PortsList[i].pPortName, 0, 1 ) ) {

			if ( PortsList[i].fPortType & PORT_TYPE_NET_ATTACHED ) {
				if ( Type & 1 )
					count++;
			}
			else {
				if ( Type & 2 )
					count++;
			}
		}

	FreePortsList();
	Sprintf( inString, L"%d", count );
	return 0;
}



DLLExports int WINAPI f_numprinters( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_COMMAS );


	wchar_t PrinterName[SHORT_BUF_LEN] = L"*";
	BOOL IPOnly = FALSE;
	int count = 0, Type = 0;

	if ( arglen[0] ) {
		for ( int i = 0; arg[0][i] != '\0'; i++ ) {

			switch ( towupper( arg[0][i] ) ) {

				case 'L' :
					Type |= PRINTER_ENUM_LOCAL;
					break;

				case 'N' :
					Type |= PRINTER_ENUM_CONNECTIONS;
					break;

				case 'I' :
					IPOnly = TRUE;
					break;

				default :
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Invalid type \"%s\"\r\n", arg[0] );
					return -1;

			}
		}
	}

	if ( IPOnly )
		Type = PRINTER_ENUM_LOCAL;

	if ( !Type )
		Type = PRINTER_ENUM_LOCAL | PRINTER_ENUM_CONNECTIONS;

	if ( arglen[1] )
		wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[1] );


	FreePrintersList();
	FreePortsList();
	PRINTER_INFO_2 * Info2 = NULL;

	if ( GetPrintersList( 2, Type ) ) {
		wcscpy_s( inString, SHORT_BUF_LEN, ErrorString );
		FreePrintersList();
		return 0;
	}

	if ( !NumPrinters ) {
		wcscpy_s( inString, SHORT_BUF_LEN, L"0" );
		FreePrintersList();
		return 0;
	}
	
	if ( IPOnly )
		if ( GetPortsList() ) {
			wcscpy_s( inString, SHORT_BUF_LEN, ErrorString );
			FreePrintersList();
			return 0;
		}

	Info2 = (PRINTER_INFO_2*) PrintersList;

	for ( unsigned int i = 0; i < NumPrinters; i++ )
		if ( !WildcardComparison( PrinterName, Info2[i].pPrinterName, 0, 1 ) ) {
			if ( !IPOnly )
				count++;
			else if ( IsIPPort( Info2[i].pPortName ) )
				count++;
		}

	FreePrintersList();
	FreePortsList();
	Sprintf( inString, L"%d", count );
	return 0;
}



DLLExports int WINAPI f_ptrcomment( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_COMMAS );


	wchar_t PrinterName[SHORT_BUF_LEN] = L"";

	if ( arglen[0] == 0 ) {

		DWORD MaxSize = SHORT_BUF_LEN;
		GetDefaultPrinter( PrinterName, &MaxSize );

		if ( PrinterName[0] == '\0' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default printer\r\n" );
			return -1;
		}
	}
	else
		wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[0] );

	FreePrinterInfo();
	PRINTER_INFO_2 * Info2;

	if ( GetPrinterInfo( PrinterName, 2 ) )
		wcscpy_s( inString, SHORT_BUF_LEN, ErrorString );
	else {
		Info2 = (PRINTER_INFO_2*) PrinterInfo;
		
		if ( Info2->pComment == NULL )
			wcscpy_s( inString, SHORT_BUF_LEN, NadaString );
		else
			wcscpy_s( inString, SHORT_BUF_LEN, Info2->pComment );

		if ( inString[0] == '\0' )
			wcscpy_s( inString, SHORT_BUF_LEN, NadaString );
	}

	FreePrinterInfo();
	return 0;
}



DLLExports int WINAPI f_ptrdriver( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_COMMAS );


	wchar_t PrinterName[SHORT_BUF_LEN] = L"";

	if ( arglen[0] == 0 ) {

		DWORD MaxSize = SHORT_BUF_LEN;
		GetDefaultPrinter( PrinterName, &MaxSize );

		if ( PrinterName[0] == '\0' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default printer\r\n" );
			return -1;
		}
	}
	else
		wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[0] );

	FreePrinterInfo();
	PRINTER_INFO_2 * Info2;

	if ( GetPrinterInfo( PrinterName, 2 ) )
		wcscpy_s( inString, SHORT_BUF_LEN, ErrorString );
	else {
		Info2 = (PRINTER_INFO_2*) PrinterInfo;

		if ( Info2->pDriverName == NULL )
			wcscpy_s( inString, SHORT_BUF_LEN, NadaString );
		else
			wcscpy_s( inString, SHORT_BUF_LEN, Info2->pDriverName );

		if ( inString[0] == '\0' )
			wcscpy_s( inString, SHORT_BUF_LEN, NadaString );
	}

	FreePrinterInfo();
	return 0;
}



int GetPrinterAddress( LPTSTR PrinterName, int Info, LPTSTR outString )
{
	if ( !PrinterName || !outString )
		return -1;

	FreePrinterInfo();
	FreePortsList();

	PRINTER_INFO_2 * Info2;


	if ( GetPrinterInfo( PrinterName, 2 ) )
		return -1;

	Info2 = (PRINTER_INFO_2*) PrinterInfo;

	if ( GetPortsList() ) {
		FreePrinterInfo();
		return -1;
	}

	if ( Info2->pServerName == NULL ) {

		//  If the printer is locally controlled, check that it really is
		//  an IP printer.  (Skip this check if it's remote.)

		if ( !IsIPPort( Info2->pPortName ) ) {
			FreePrinterInfo();
			FreePortsList();
			return -2;
		}
	}


	GetPortAddress( Info2->pPortName, Info2->pServerName, Info, outString );

	FreePrinterInfo();
	FreePortsList();
	return 0;
}



DLLExports int WINAPI f_ptrexist( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_COMMAS );


	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing printer name\r\n" );
		return -1;
	}

	if ( PrinterExists( arg[0] ) )
		wcscpy_s( inString, MAX_ARG_LEN, L"1" );
	else
		wcscpy_s( inString, MAX_ARG_LEN, L"0" );

	return 0;
}



DLLExports int WINAPI f_ptrhost( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_COMMAS );


	wchar_t PrinterName[SHORT_BUF_LEN] = L"";

	if ( arglen[0] == 0 ) {

		DWORD MaxSize = SHORT_BUF_LEN;
		GetDefaultPrinter( PrinterName, &MaxSize );

		if ( PrinterName[0] == '\0' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default printer\r\n" );
			return -1;
		}
	}
	else
		wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[0] );

	int rv = GetPrinterAddress( PrinterName, 1, inString );

	if ( rv == -2 )
		wcscpy_s( inString, SHORT_BUF_LEN, L"Not an IP printer" );
	else if ( rv < 0 )
		wcscpy_s( inString, SHORT_BUF_LEN, ErrorString );

	FreePrinterInfo();
	FreePortsList();
	return 0;
}



DLLExports int WINAPI f_ptrip( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_COMMAS );


	wchar_t PrinterName[SHORT_BUF_LEN] = L"";

	if ( arglen[0] == 0 ) {

		DWORD MaxSize = SHORT_BUF_LEN;
		GetDefaultPrinter( PrinterName, &MaxSize );

		if ( PrinterName[0] == '\0' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default printer\r\n" );
			return -1;
		}
	}
	else
		wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[0] );

	int rv = GetPrinterAddress( PrinterName, 0, inString );

	if ( rv == -2 )
		wcscpy_s( inString, SHORT_BUF_LEN, L"Not an IP printer" );
	else if ( rv < 0 )
		wcscpy_s( inString, SHORT_BUF_LEN, ErrorString );

	FreePrinterInfo();
	FreePortsList();
	return 0;
}



DLLExports int WINAPI f_ptrjobs( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_COMMAS );


	wchar_t PrinterName[SHORT_BUF_LEN] = L"";

	if ( arglen[0] == 0 ) {

		DWORD MaxSize = SHORT_BUF_LEN;
		GetDefaultPrinter( PrinterName, &MaxSize );

		if ( PrinterName[0] == '\0' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default printer\r\n" );
			return -1;
		}
	}
	else
		wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[0] );

	FreePrinterInfo();
	PRINTER_INFO_2 * Info2;

	if ( GetPrinterInfo( PrinterName, 2 ) )
		wcscpy_s( inString, SHORT_BUF_LEN, ErrorString );
	else {
		Info2 = (PRINTER_INFO_2*) PrinterInfo;

		Sprintf( inString, L"%d", Info2->cJobs );
	}

	FreePrinterInfo();
	return 0;
}



DLLExports int WINAPI f_ptrloc( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_COMMAS );


	wchar_t PrinterName[SHORT_BUF_LEN] = L"";

	if ( arglen[0] == 0 ) {

		DWORD MaxSize = SHORT_BUF_LEN;
		GetDefaultPrinter( PrinterName, &MaxSize );

		if ( PrinterName[0] == '\0' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default printer\r\n" );
			return -1;
		}
	}
	else
		wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[0] );

	FreePrinterInfo();
	PRINTER_INFO_2 * Info2;

	if ( GetPrinterInfo( PrinterName, 2 ) )
		wcscpy_s( inString, SHORT_BUF_LEN, ErrorString );
	else {
		Info2 = (PRINTER_INFO_2*) PrinterInfo;

		if ( Info2->pLocation == NULL )
			wcscpy_s( inString, SHORT_BUF_LEN, NadaString );
		else
			wcscpy_s( inString, SHORT_BUF_LEN, Info2->pLocation );

		if ( inString[0] == '\0' )
			wcscpy_s( inString, SHORT_BUF_LEN, NadaString );
	}

	FreePrinterInfo();
	return 0;
}



DLLExports int WINAPI f_ptrname( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_COMMAS );


	int Type = 0;
	unsigned int index = -1;

	if ( arglen[0] == 0 ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Missing index\r\n" );
		return -1;
	}
	else if ( !iswdigit( arg[0][0] ) ) {
		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"Invalid index \"%s\"\r\n", arg[0] );
		return -1;
	}
	else {
		index = GetInt( arg[0] );

		if ( index < 0 ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Invalid index \"%s\"\r\n", arg[0] );
			return -1;
		}
	}

	if ( arglen[1] ) {
		for ( int i = 0; arg[1][i] != '\0'; i++ ) {

			switch ( towupper( arg[1][i] ) ) {

				case 'L' :
					Type |= PRINTER_ENUM_LOCAL;
					break;

				case 'N' :
					Type |= PRINTER_ENUM_CONNECTIONS;
					break;

				default :
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Invalid type \"%s\"\r\n", arg[1] );
					return -1;

			}
		}
	}


	if ( !Type )
		Type = PRINTER_ENUM_LOCAL | PRINTER_ENUM_CONNECTIONS;

	FreePrintersList();
	PRINTER_INFO_2 * Info2 = NULL;

	if ( GetPrintersList( 2, Type ) ) {
		wcscpy_s( inString, SHORT_BUF_LEN, ErrorString );
		return -1;
	}

	Info2 = (PRINTER_INFO_2*) PrintersList;

	if ( index >= NumPrinters )
		inString[0] = '\0';
	else
		wcscpy_s( inString, SHORT_BUF_LEN, Info2[index].pPrinterName );

	FreePrintersList();
	return 0;
}



DLLExports int WINAPI f_ptroffline( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_COMMAS );


	wchar_t PrinterName[SHORT_BUF_LEN] = L"";
	int rv = 0;

	if ( arglen[0] == 0 ) {

		DWORD MaxSize = SHORT_BUF_LEN;
		GetDefaultPrinter( PrinterName, &MaxSize );

		if ( PrinterName[0] == '\0' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default printer\r\n" );
			return -1;
		}
	}
	else
		wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[0] );

	FreePrinterInfo();
	PRINTER_INFO_2 * Info2;

	if ( GetPrinterInfo( PrinterName, 2 ) )
		wcscpy_s( inString, SHORT_BUF_LEN, ErrorString );
	else {
		Info2 = (PRINTER_INFO_2*) PrinterInfo;

		if ( Info2->Attributes & PRINTER_ATTRIBUTE_WORK_OFFLINE )
			rv = 1;
		else if ( Info2->Status & PRINTER_STATUS_OFFLINE )
			rv = 1;

		Sprintf( inString, L"%d", rv );
	}

	return 0;
}



DLLExports int WINAPI f_ptrport( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_COMMAS );


	wchar_t PrinterName[SHORT_BUF_LEN] = L"";

	if ( arglen[0] == 0 ) {

		DWORD MaxSize = SHORT_BUF_LEN;
		GetDefaultPrinter( PrinterName, &MaxSize );

		if ( PrinterName[0] == '\0' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default printer\r\n" );
			return -1;
		}
	}
	else
		wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[0] );

	FreePrinterInfo();
	PRINTER_INFO_2 * Info2;

	if ( GetPrinterInfo( PrinterName, 2 ) )
		wcscpy_s( inString, SHORT_BUF_LEN, ErrorString );
	else {
		Info2 = (PRINTER_INFO_2*) PrinterInfo;

		if ( Info2->pPortName == NULL )
			wcscpy_s( inString, SHORT_BUF_LEN, NadaString );
		else
			wcscpy_s( inString, SHORT_BUF_LEN, Info2->pPortName );

		if ( inString[0] == '\0' )
			wcscpy_s( inString, SHORT_BUF_LEN, NadaString );
	}

	FreePrinterInfo();
	return 0;
}



DLLExports int WINAPI f_ptrserver( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_COMMAS );


	wchar_t PrinterName[SHORT_BUF_LEN] = L"";

	if ( arglen[0] == 0 ) {

		DWORD MaxSize = SHORT_BUF_LEN;
		GetDefaultPrinter( PrinterName, &MaxSize );

		if ( PrinterName[0] == '\0' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default printer\r\n" );
			return -1;
		}
	}
	else
		wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[0] );

	FreePrinterInfo();
	PRINTER_INFO_2 * Info2;

	if ( GetPrinterInfo( PrinterName, 2 ) )
		wcscpy_s( inString, SHORT_BUF_LEN, ErrorString );
	else {
		Info2 = (PRINTER_INFO_2*) PrinterInfo;

		if ( Info2->pServerName == NULL )
			wcscpy_s( inString, SHORT_BUF_LEN, NadaString );
		else
			wcscpy_s( inString, SHORT_BUF_LEN, Info2->pServerName );

		if ( inString[0] == '\0' )
			wcscpy_s( inString, SHORT_BUF_LEN, NadaString );
	}

	FreePrinterInfo();
	return 0;
}



DLLExports int WINAPI f_ptrshare( LPTSTR inString )
{
	if ( !inString )
		return 1;


	ParseArgs( inString, PARSE_BREAK_COMMAS );


	wchar_t PrinterName[SHORT_BUF_LEN] = L"";

	if ( arglen[0] == 0 ) {

		DWORD MaxSize = SHORT_BUF_LEN;
		GetDefaultPrinter( PrinterName, &MaxSize );

		if ( PrinterName[0] == '\0' ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"No default printer\r\n" );
			return -1;
		}
	}
	else
		wcscpy_s( PrinterName, SHORT_BUF_LEN, arg[0] );

	FreePrinterInfo();
	PRINTER_INFO_2 * Info2;

	if ( GetPrinterInfo( PrinterName, 2 ) )
		wcscpy_s( inString, SHORT_BUF_LEN, ErrorString );
	else {
		Info2 = (PRINTER_INFO_2*) PrinterInfo;

		if ( Info2->pShareName == NULL )
			wcscpy_s( inString, SHORT_BUF_LEN, NadaString );
		else
			wcscpy_s( inString, SHORT_BUF_LEN, Info2->pShareName );

		if ( inString[0] == '\0' )
			wcscpy_s( inString, SHORT_BUF_LEN, NadaString );
	}

	FreePrinterInfo();
	return 0;
}



DLLExports int WINAPI ptrstuffhelp( LPTSTR inString )
{
	return PluginHelpCmd( inString, PtrStuffHelpHelpText );
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/printerstuff.html";
	piInfo.pszDescription	= (LPTSTR)L"Create, delete, and manage printers";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( !wcscmp( shortbuf, L"00" ) )
		Printf( L"%s v%d.%d.%d loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{
	FreePortsList();
	FreePrintersList();
	CloseHelpWindow();
	return 0;
}



DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}
